# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

import awkward.forth


def test_basics():
    vm32 = awkward.forth.ForthMachine32("")
    vm64 = awkward.forth.ForthMachine64("")

    assert str(ak.type(vm32.bytecodes)) == "var * int32"
    assert str(ak.type(vm64.bytecodes)) == "var * int32"
    assert vm32.dictionary == []
    assert vm64.dictionary == []
    assert vm32.stack_max_depth == 1024
    assert vm64.stack_max_depth == 1024
    assert vm32.recursion_max_depth == 1024
    assert vm64.recursion_max_depth == 1024
    assert vm32.output_initial_size == 1024
    assert vm64.output_initial_size == 1024
    assert vm32.output_resize_factor == 1.5
    assert vm64.output_resize_factor == 1.5

    vm32.stack_push(1)
    vm32.stack_push(2)
    vm32.stack_push(3)
    assert vm32.stack_pop() == 3
    assert vm32.stack == [1, 2]
    vm32.stack_clear()
    assert vm32.stack == []
    vm64.stack_push(1)
    vm64.stack_push(2)
    vm64.stack_push(3)
    assert vm64.stack_pop() == 3
    assert vm64.stack == [1, 2]
    vm64.stack_clear()
    assert vm64.stack == []

    assert vm32.variables == {}
    assert vm64.variables == {}
    assert vm32.outputs == {}
    assert vm64.outputs == {}


def test_comment_compilation():
    vm32 = awkward.forth.ForthMachine32("")
    assert ak.to_list(vm32.bytecodes) == [[]]
    assert vm32.decompiled == ""

    vm32 = awkward.forth.ForthMachine32("( comment )")
    assert ak.to_list(vm32.bytecodes) == [[]]

    vm32 = awkward.forth.ForthMachine32("\\ comment")
    assert ak.to_list(vm32.bytecodes) == [[]]

    vm32 = awkward.forth.ForthMachine32("\\ comment\n")
    assert ak.to_list(vm32.bytecodes) == [[]]

    vm32 = awkward.forth.ForthMachine32("1 2 ( comment ) 3 4")
    assert ak.to_list(vm32.bytecodes) == [[0, 1, 0, 2, 0, 3, 0, 4]]

    vm32 = awkward.forth.ForthMachine32("1 2 \\ comment \n 3 4")
    assert ak.to_list(vm32.bytecodes) == [[0, 1, 0, 2, 0, 3, 0, 4]]


def test_literal_compilation():
    vm32 = awkward.forth.ForthMachine32("1 2 3 4")
    assert ak.to_list(vm32.bytecodes) == [[0, 1, 0, 2, 0, 3, 0, 4]]
    assert (
        vm32.decompiled
        == """1
2
3
4
"""
    )


def test_userdef_compilation():
    vm32 = awkward.forth.ForthMachine32(": stuff 1 2 3 4 ;")
    assert ak.to_list(vm32.bytecodes) == [[], [0, 1, 0, 2, 0, 3, 0, 4]]
    assert (
        vm32.decompiled
        == """: stuff
  1
  2
  3
  4
;
"""
    )

    vm32 = awkward.forth.ForthMachine32(": foo 123 : bar 1 2 3 ; 321 ;")
    assert ak.to_list(vm32.bytecodes) == [[], [0, 123, 0, 321], [0, 1, 0, 2, 0, 3]]
    assert (
        vm32.decompiled
        == """: foo
  123
  321
;

: bar
  1
  2
  3
;
"""
    )

    vm32 = awkward.forth.ForthMachine32(": empty ;")
    assert ak.to_list(vm32.bytecodes) == [[], []]
    assert (
        vm32.decompiled
        == """: empty
;
"""
    )

    vm32 = awkward.forth.ForthMachine32(": infinite recurse ;")
    assert (
        vm32.decompiled
        == """: infinite
  infinite
;
"""
    )


def test_declarations_compilation():
    vm32 = awkward.forth.ForthMachine32("variable x")
    assert ak.to_list(vm32.bytecodes) == [[]]
    assert (
        vm32.decompiled
        == """variable x
"""
    )

    vm32 = awkward.forth.ForthMachine32("variable x variable y")
    assert ak.to_list(vm32.bytecodes) == [[]]
    assert (
        vm32.decompiled
        == """variable x
variable y
"""
    )

    vm32 = awkward.forth.ForthMachine32("input x")
    assert ak.to_list(vm32.bytecodes) == [[]]
    assert (
        vm32.decompiled
        == """input x
"""
    )

    vm32 = awkward.forth.ForthMachine32("output x int32")
    assert ak.to_list(vm32.bytecodes) == [[]]
    assert (
        vm32.decompiled
        == """output x int32
"""
    )

    vm32 = awkward.forth.ForthMachine32("output x int32 input y variable z")
    assert ak.to_list(vm32.bytecodes) == [[]]
    assert (
        vm32.decompiled
        == """variable z
input y
output x int32
"""
    )

    for dtype in [
        "bool",
        "int8",
        "int16",
        "int32",
        "int64",
        "uint8",
        "uint16",
        "uint32",
        "uint64",
        "float32",
        "float64",
    ]:
        vm32 = awkward.forth.ForthMachine32("output x " + dtype)
        assert vm32.decompiled == "output x " + dtype + "\n"


def test_control_compilation():
    vm32 = awkward.forth.ForthMachine32("halt")
    assert (
        vm32.decompiled
        == """halt
"""
    )

    vm32 = awkward.forth.ForthMachine32("pause")
    assert (
        vm32.decompiled
        == """pause
"""
    )

    vm32 = awkward.forth.ForthMachine32("exit")
    assert (
        vm32.decompiled
        == """exit
"""
    )


def test_if_compilation():
    vm32 = awkward.forth.ForthMachine32("-1 if then")
    assert (
        vm32.decompiled
        == """-1
if
then
"""
    )

    vm32 = awkward.forth.ForthMachine32("-1 if 3 5 + then")
    assert (
        vm32.decompiled
        == """-1
if
  3
  5
  +
then
"""
    )

    vm32 = awkward.forth.ForthMachine32("-1 if else then")
    assert (
        vm32.decompiled
        == """-1
if
else
then
"""
    )

    vm32 = awkward.forth.ForthMachine32("-1 if 3 5 + else 123 then")
    assert (
        vm32.decompiled
        == """-1
if
  3
  5
  +
else
  123
then
"""
    )

    vm32 = awkward.forth.ForthMachine32(": foo -1 if 3 5 + then ; foo")
    assert (
        vm32.decompiled
        == """: foo
  -1
  if
    3
    5
    +
  then
;

foo
"""
    )


def test_loop_compilation():
    vm32 = awkward.forth.ForthMachine32("10 0 do loop")
    assert (
        vm32.decompiled
        == """10
0
do
loop
"""
    )

    vm32 = awkward.forth.ForthMachine32("10 0 do i loop")
    assert (
        vm32.decompiled
        == """10
0
do
  i
loop
"""
    )

    vm32 = awkward.forth.ForthMachine32("10 0 do i +loop")
    assert (
        vm32.decompiled
        == """10
0
do
  i
+loop
"""
    )

    vm32 = awkward.forth.ForthMachine32("10 0 do 5 0 do 3 1 do i j k loop loop loop")
    assert (
        vm32.decompiled
        == """10
0
do
  5
  0
  do
    3
    1
    do
      i
      j
      k
    loop
  loop
loop
"""
    )

    vm32 = awkward.forth.ForthMachine32("begin again")
    assert (
        vm32.decompiled
        == """begin
again
"""
    )

    vm32 = awkward.forth.ForthMachine32("begin 123 again")
    assert (
        vm32.decompiled
        == """begin
  123
again
"""
    )

    vm32 = awkward.forth.ForthMachine32("begin until")
    assert (
        vm32.decompiled
        == """begin
until
"""
    )

    vm32 = awkward.forth.ForthMachine32("begin 123 until")
    assert (
        vm32.decompiled
        == """begin
  123
until
"""
    )

    vm32 = awkward.forth.ForthMachine32("begin while repeat")
    assert (
        vm32.decompiled
        == """begin
while
repeat
"""
    )

    vm32 = awkward.forth.ForthMachine32("begin 123 while 321 repeat")
    assert (
        vm32.decompiled
        == """begin
  123
while
  321
repeat
"""
    )


def test_io_compilation():
    vm32 = awkward.forth.ForthMachine32("variable x x !")
    assert (
        vm32.decompiled
        == """variable x

x !
"""
    )

    vm32 = awkward.forth.ForthMachine32("variable x x +!")
    assert (
        vm32.decompiled
        == """variable x

x +!
"""
    )

    vm32 = awkward.forth.ForthMachine32("variable x x @")
    assert (
        vm32.decompiled
        == """variable x

x @
"""
    )

    vm32 = awkward.forth.ForthMachine32("input x x len")
    assert (
        vm32.decompiled
        == """input x

x len
"""
    )

    vm32 = awkward.forth.ForthMachine32("input x x pos")
    assert (
        vm32.decompiled
        == """input x

x pos
"""
    )

    vm32 = awkward.forth.ForthMachine32("input x x end")
    assert (
        vm32.decompiled
        == """input x

x end
"""
    )

    vm32 = awkward.forth.ForthMachine32("input x x seek")
    assert (
        vm32.decompiled
        == """input x

x seek
"""
    )

    vm32 = awkward.forth.ForthMachine32("input x x skip")
    assert (
        vm32.decompiled
        == """input x

x skip
"""
    )

    vm32 = awkward.forth.ForthMachine32("output x int32 x len")
    assert (
        vm32.decompiled
        == """output x int32

x len
"""
    )

    vm32 = awkward.forth.ForthMachine32("output x int32 x rewind")
    assert (
        vm32.decompiled
        == """output x int32

x rewind
"""
    )


def test_read_compilation():
    vm32 = awkward.forth.ForthMachine32("input x x i-> stack")
    assert (
        vm32.decompiled
        == """input x

x i-> stack
"""
    )

    vm32 = awkward.forth.ForthMachine32("input x output y int32 x i-> y")
    assert (
        vm32.decompiled
        == """input x
output y int32

x i-> y
"""
    )

    for rep in ["", "#"]:
        for big in ["", "!"]:
            for tpe in [
                "?",
                "b",
                "h",
                "i",
                "q",
                "n",
                "B",
                "H",
                "I",
                "Q",
                "N",
                "f",
                "d",
            ]:
                if not (big == "!" and tpe in ("?", "b", "B")):
                    source = """input x

x {0} stack
""".format(
                        rep + big + tpe + "->"
                    )
                    vm32 = awkward.forth.ForthMachine32(source)
                    assert vm32.decompiled == source
                    source = """input x
output y int32

x {0} y
""".format(
                        rep + big + tpe + "->"
                    )
                    vm32 = awkward.forth.ForthMachine32(source)
                    assert vm32.decompiled == source
                    del vm32


def test_everything_else_compilation():
    source = (
        "\n".join(
            [
                "dup",
                "drop",
                "swap",
                "over",
                "rot",
                "nip",
                "tuck",
                "+",
                "-",
                "*",
                "/",
                "mod",
                "/mod",
                "negate",
                "1+",
                "1-",
                "abs",
                "min",
                "max",
                "=",
                "<>",
                ">",
                ">=",
                "<",
                "<=",
                "0=",
                "invert",
                "and",
                "or",
                "xor",
                "lshift",
                "rshift",
                "false",
                "true",
            ]
        )
        + "\n"
    )
    vm32 = awkward.forth.ForthMachine32(source)
    assert vm32.decompiled == source


def test_input_output_type():
    vm32 = awkward.forth.ForthMachine32("input x output y int32")
    vm32.begin({"x": np.array([1, 2, 3])})
    assert isinstance(vm32["y"], ak.layout.NumpyArray)


def test_stepping():
    vm32 = awkward.forth.ForthMachine32("1 2 3 4")
    with pytest.raises(ValueError):
        vm32.step()
    assert not vm32.is_ready
    assert vm32.is_done

    vm32.begin()
    assert vm32.is_ready
    assert not vm32.is_done
    assert vm32.stack == []
    vm32.step()
    assert vm32.stack == [1]
    vm32.step()
    assert vm32.stack == [1, 2]
    vm32.step()
    assert vm32.stack == [1, 2, 3]
    vm32.step()
    assert vm32.stack == [1, 2, 3, 4]
    assert vm32.is_ready
    assert vm32.is_done
    with pytest.raises(ValueError):
        vm32.step()

    vm32.reset()
    with pytest.raises(ValueError):
        vm32.step()

    vm32.begin()
    assert vm32.stack == []
    vm32.step()
    assert vm32.stack == [1]
    vm32.step()
    assert vm32.stack == [1, 2]
    vm32.step()
    assert vm32.stack == [1, 2, 3]
    vm32.step()
    assert vm32.stack == [1, 2, 3, 4]
    with pytest.raises(ValueError):
        vm32.step()


def test_running():
    vm32 = awkward.forth.ForthMachine32("1 2 3 4")
    vm32.run()
    assert vm32.is_ready
    assert vm32.is_done

    assert vm32.stack == [1, 2, 3, 4]
    with pytest.raises(ValueError):
        vm32.step()

    vm32.run()

    assert vm32.stack == [1, 2, 3, 4]
    with pytest.raises(ValueError):
        vm32.step()


def test_pausing():
    vm32 = awkward.forth.ForthMachine32("1 2 pause 3 4 5")

    vm32.run()
    assert vm32.is_ready
    assert not vm32.is_done
    assert vm32.stack == [1, 2]
    vm32.step()
    assert vm32.is_ready
    assert not vm32.is_done
    assert vm32.stack == [1, 2, 3]
    vm32.resume()
    assert vm32.is_ready
    assert vm32.is_done
    assert vm32.stack == [1, 2, 3, 4, 5]


def test_calling():
    vm32 = awkward.forth.ForthMachine32(": foo 999 ; 1 2 3 pause 4 5")

    vm32.begin()
    assert vm32.stack == []
    assert not vm32.is_done

    vm32.call("foo")
    assert vm32.stack == [999]
    assert not vm32.is_done

    vm32.step()
    assert vm32.stack == [999, 1]
    assert not vm32.is_done

    vm32.call("foo")
    assert vm32.stack == [999, 1, 999]
    assert not vm32.is_done

    vm32.resume()
    assert vm32.stack == [999, 1, 999, 2, 3]
    assert not vm32.is_done

    vm32.resume()
    assert vm32.stack == [999, 1, 999, 2, 3, 4, 5]
    assert vm32.is_done

    vm32.run()
    assert vm32.stack == [1, 2, 3]
    assert not vm32.is_done

    vm32.call("foo")
    assert vm32.stack == [1, 2, 3, 999]
    assert not vm32.is_done

    vm32.call("foo")
    assert vm32.stack == [1, 2, 3, 999, 999]
    assert not vm32.is_done

    vm32.step()
    assert vm32.stack == [1, 2, 3, 999, 999, 4]
    assert not vm32.is_done

    vm32.step()
    assert vm32.stack == [1, 2, 3, 999, 999, 4, 5]
    assert vm32.is_done


def test_calling2():
    vm32 = awkward.forth.ForthMachine32(
        """
: bar 999 ;
: foo bar pause bar ;
1 2 3 pause 4 5"""
    )

    vm32.run()
    assert vm32.stack == [1, 2, 3]
    assert not vm32.is_done

    vm32.call("foo")
    assert vm32.stack == [1, 2, 3, 999]
    assert not vm32.is_done

    vm32.resume()
    assert vm32.stack == [1, 2, 3, 999, 999]
    assert not vm32.is_done

    vm32.resume()
    assert vm32.stack == [1, 2, 3, 999, 999, 4, 5]
    assert vm32.is_done

    vm32.call("foo")
    assert vm32.stack == [1, 2, 3, 999, 999, 4, 5, 999]
    assert not vm32.is_done

    vm32.resume()
    assert vm32.stack == [1, 2, 3, 999, 999, 4, 5, 999, 999]
    assert vm32.is_done


def test_calling3():
    vm32 = awkward.forth.ForthMachine32(
        """
: bar 999 ;
: foo bar pause bar ;"""
    )

    vm32.run()
    assert vm32.stack == []
    assert vm32.is_done

    vm32.call("foo")
    assert vm32.stack == [999]
    assert not vm32.is_done

    vm32.resume()
    assert vm32.stack == [999, 999]
    assert vm32.is_done


def test_halt():
    vm32 = awkward.forth.ForthMachine32("1 2 3 halt 4 5")

    vm32.run(raise_user_halt=False)
    assert vm32.stack == [1, 2, 3]
    assert vm32.is_done


def test_halt2():
    vm32 = awkward.forth.ForthMachine32(
        """
: bar 999 ;
: foo bar halt bar ;
1 2 3 pause 4 5"""
    )

    vm32.run()
    assert vm32.stack == [1, 2, 3]
    assert not vm32.is_done

    vm32.call("foo", raise_user_halt=False)
    assert vm32.stack == [1, 2, 3, 999]
    assert vm32.is_done

    with pytest.raises(ValueError):
        vm32.step()

    vm32.run()
    assert vm32.stack == [1, 2, 3]
    assert not vm32.is_done

    vm32.resume()
    assert vm32.stack == [1, 2, 3, 4, 5]
    assert vm32.is_done


def test_do():
    vm32 = awkward.forth.ForthMachine32("5 0 do i loop")
    vm32.run()
    assert vm32.stack == [0, 1, 2, 3, 4]

    vm32 = awkward.forth.ForthMachine32("5 0 do i pause loop")
    vm32.run()
    assert vm32.stack == [0]
    vm32.resume()
    assert vm32.stack == [0, 1]
    vm32.resume()
    assert vm32.stack == [0, 1, 2]
    vm32.resume()
    assert vm32.stack == [0, 1, 2, 3]
    vm32.resume()
    assert vm32.stack == [0, 1, 2, 3, 4]

    vm32 = awkward.forth.ForthMachine32(": foo 123 pause ; 5 0 do i foo loop")
    vm32.run()
    assert vm32.stack == [0, 123]
    vm32.resume()
    assert vm32.stack == [0, 123, 1, 123]
    vm32.resume()
    assert vm32.stack == [0, 123, 1, 123, 2, 123]
    vm32.resume()
    assert vm32.stack == [0, 123, 1, 123, 2, 123, 3, 123]
    vm32.resume()
    assert vm32.stack == [0, 123, 1, 123, 2, 123, 3, 123, 4, 123]

    vm32 = awkward.forth.ForthMachine32(": foo 123 pause ; 5 0 do i foo loop")
    vm32.run()
    assert vm32.stack == [0, 123]
    vm32.call("foo")
    assert vm32.stack == [0, 123, 123]
    vm32.resume()
    assert vm32.stack == [0, 123, 123, 1, 123]


def test_errors():
    # util::ForthError::not_ready
    vm32 = awkward.forth.ForthMachine32("1 2 3")
    with pytest.raises(ValueError):
        vm32.resume()
    with pytest.raises(ValueError):
        vm32.step()

    # util::ForthError::is_done
    vm32 = awkward.forth.ForthMachine32("1 2 3")
    vm32.run()
    with pytest.raises(ValueError):
        vm32.resume()
    with pytest.raises(ValueError):
        vm32.step()

    # util::ForthError::user_halt
    vm32 = awkward.forth.ForthMachine32("halt")
    with pytest.raises(ValueError):
        vm32.run()
    vm32.run(raise_user_halt=False)

    # util::ForthError::recursion_depth_exceeded
    vm32 = awkward.forth.ForthMachine32(": infinite infinite ; infinite")
    with pytest.raises(ValueError):
        vm32.run()
    vm32.run(raise_recursion_depth_exceeded=False)

    # util::ForthError::stack_underflow
    vm32 = awkward.forth.ForthMachine32("1 +")
    with pytest.raises(ValueError):
        vm32.run()
    vm32.run(raise_stack_underflow=False)

    # util::ForthError::stack_overflow
    vm32 = awkward.forth.ForthMachine32("1025 0 do i loop")
    with pytest.raises(ValueError):
        vm32.run()
    vm32.run(raise_stack_overflow=False)

    # util::ForthError::read_beyond
    vm32 = awkward.forth.ForthMachine32("input x x b-> stack")
    with pytest.raises(ValueError):
        vm32.run({"x": np.array([])})
    vm32.run({"x": np.array([])}, raise_read_beyond=False)

    # util::ForthError::seek_beyond
    vm32 = awkward.forth.ForthMachine32("input x 1 x seek")
    with pytest.raises(ValueError):
        vm32.run({"x": np.array([])})
    vm32.run({"x": np.array([])}, raise_seek_beyond=False)

    # util::ForthError::skip_beyond
    vm32 = awkward.forth.ForthMachine32("input x 1 x skip")
    with pytest.raises(ValueError):
        vm32.run({"x": np.array([])})
    vm32.run({"x": np.array([])}, raise_skip_beyond=False)

    # util::ForthError::rewind_beyond
    vm32 = awkward.forth.ForthMachine32("output x int32 1 x rewind")
    with pytest.raises(ValueError):
        vm32.run()
    vm32.run(raise_rewind_beyond=False)

    # util::ForthError::division_by_zero
    vm32 = awkward.forth.ForthMachine32("123 0 /")
    with pytest.raises(ValueError):
        vm32.run()
    vm32.run(raise_division_by_zero=False)

    vm32 = awkward.forth.ForthMachine32("-123 0 /")
    with pytest.raises(ValueError):
        vm32.run()
    vm32.run(raise_division_by_zero=False)

    vm32 = awkward.forth.ForthMachine32("123 0 mod")
    with pytest.raises(ValueError):
        vm32.run()
    vm32.run(raise_division_by_zero=False)

    vm32 = awkward.forth.ForthMachine32("-123 0 mod")
    with pytest.raises(ValueError):
        vm32.run()
    vm32.run(raise_division_by_zero=False)

    vm32 = awkward.forth.ForthMachine32("123 0 /mod")
    with pytest.raises(ValueError):
        vm32.run()
    vm32.run(raise_division_by_zero=False)

    vm32 = awkward.forth.ForthMachine32("-123 0 /mod")
    with pytest.raises(ValueError):
        vm32.run()
    vm32.run(raise_division_by_zero=False)


def test_gforth():
    # Expected outputs were all generated by running gforth:
    #
    #     echo "100 maxdepth-.s ! {0} .s" | gforth
    #
    # and parsing the stack output (none of the stacks are larger than 100).

    vm32 = awkward.forth.ForthMachine32(": foo if 999 then ; -1 foo")
    vm32.run()
    assert vm32.stack == [999]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 then ; 0 foo")
    vm32.run()
    assert vm32.stack == []

    vm32 = awkward.forth.ForthMachine32(": foo if 999 then ; 1 foo")
    vm32.run()
    assert vm32.stack == [999]

    vm32 = awkward.forth.ForthMachine32(": foo if if 999 then then ; -1 -1 foo")
    vm32.run()
    assert vm32.stack == [999]

    vm32 = awkward.forth.ForthMachine32(": foo if if 999 then then ; 0 -1 foo")
    vm32.run()
    assert vm32.stack == []

    vm32 = awkward.forth.ForthMachine32(": foo if if 999 then then ; 1 -1 foo")
    vm32.run()
    assert vm32.stack == [999]

    vm32 = awkward.forth.ForthMachine32(": foo if if 999 then then ; -1 0 foo")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32(": foo if if 999 then then ; 0 0 foo")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32(": foo if if 999 then then ; 1 0 foo")
    vm32.run()
    assert vm32.stack == [1]

    vm32 = awkward.forth.ForthMachine32(": foo if if 999 then then ; -1 1 foo")
    vm32.run()
    assert vm32.stack == [999]

    vm32 = awkward.forth.ForthMachine32(": foo if if 999 then then ; 0 1 foo")
    vm32.run()
    assert vm32.stack == []

    vm32 = awkward.forth.ForthMachine32(": foo if if 999 then then ; 1 1 foo")
    vm32.run()
    assert vm32.stack == [999]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else 123 then ; -1 foo")
    vm32.run()
    assert vm32.stack == [999]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else 123 then ; 0 foo")
    vm32.run()
    assert vm32.stack == [123]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else 123 then ; 1 foo")
    vm32.run()
    assert vm32.stack == [999]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else 123 then ; -1 -1 foo")
    vm32.run()
    assert vm32.stack == [-1, 999]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else 123 then ; 0 -1 foo")
    vm32.run()
    assert vm32.stack == [0, 999]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else 123 then ; 1 -1 foo")
    vm32.run()
    assert vm32.stack == [1, 999]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else 123 then ; -1 0 foo")
    vm32.run()
    assert vm32.stack == [-1, 123]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else 123 then ; 0 0 foo")
    vm32.run()
    assert vm32.stack == [0, 123]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else 123 then ; 1 0 foo")
    vm32.run()
    assert vm32.stack == [1, 123]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else 123 then ; -1 1 foo")
    vm32.run()
    assert vm32.stack == [-1, 999]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else 123 then ; 0 1 foo")
    vm32.run()
    assert vm32.stack == [0, 999]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else 123 then ; 1 1 foo")
    vm32.run()
    assert vm32.stack == [1, 999]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if if 999 then else 123 then ; -1 -1 foo"
    )
    vm32.run()
    assert vm32.stack == [999]

    vm32 = awkward.forth.ForthMachine32(": foo if if 999 then else 123 then ; 0 -1 foo")
    vm32.run()
    assert vm32.stack == []

    vm32 = awkward.forth.ForthMachine32(": foo if if 999 then else 123 then ; 1 -1 foo")
    vm32.run()
    assert vm32.stack == [999]

    vm32 = awkward.forth.ForthMachine32(": foo if if 999 then else 123 then ; -1 0 foo")
    vm32.run()
    assert vm32.stack == [-1, 123]

    vm32 = awkward.forth.ForthMachine32(": foo if if 999 then else 123 then ; 0 0 foo")
    vm32.run()
    assert vm32.stack == [0, 123]

    vm32 = awkward.forth.ForthMachine32(": foo if if 999 then else 123 then ; 1 0 foo")
    vm32.run()
    assert vm32.stack == [1, 123]

    vm32 = awkward.forth.ForthMachine32(": foo if if 999 then else 123 then ; -1 1 foo")
    vm32.run()
    assert vm32.stack == [999]

    vm32 = awkward.forth.ForthMachine32(": foo if if 999 then else 123 then ; 0 1 foo")
    vm32.run()
    assert vm32.stack == []

    vm32 = awkward.forth.ForthMachine32(": foo if if 999 then else 123 then ; 1 1 foo")
    vm32.run()
    assert vm32.stack == [999]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if 999 else if 123 then then ; -1 -1 foo"
    )
    vm32.run()
    assert vm32.stack == [-1, 999]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else if 123 then then ; 0 -1 foo")
    vm32.run()
    assert vm32.stack == [0, 999]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else if 123 then then ; 1 -1 foo")
    vm32.run()
    assert vm32.stack == [1, 999]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else if 123 then then ; -1 0 foo")
    vm32.run()
    assert vm32.stack == [123]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else if 123 then then ; 0 0 foo")
    vm32.run()
    assert vm32.stack == []

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else if 123 then then ; 1 0 foo")
    vm32.run()
    assert vm32.stack == [123]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else if 123 then then ; -1 1 foo")
    vm32.run()
    assert vm32.stack == [-1, 999]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else if 123 then then ; 0 1 foo")
    vm32.run()
    assert vm32.stack == [0, 999]

    vm32 = awkward.forth.ForthMachine32(": foo if 999 else if 123 then then ; 1 1 foo")
    vm32.run()
    assert vm32.stack == [1, 999]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if if 999 else 321 then else 123 then ; -1 -1 foo"
    )
    vm32.run()
    assert vm32.stack == [999]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if if 999 else 321 then else 123 then ; 0 -1 foo"
    )
    vm32.run()
    assert vm32.stack == [321]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if if 999 else 321 then else 123 then ; 1 -1 foo"
    )
    vm32.run()
    assert vm32.stack == [999]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if if 999 else 321 then else 123 then ; -1 0 foo"
    )
    vm32.run()
    assert vm32.stack == [-1, 123]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if if 999 else 321 then else 123 then ; 0 0 foo"
    )
    vm32.run()
    assert vm32.stack == [0, 123]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if if 999 else 321 then else 123 then ; 1 0 foo"
    )
    vm32.run()
    assert vm32.stack == [1, 123]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if if 999 else 321 then else 123 then ; -1 1 foo"
    )
    vm32.run()
    assert vm32.stack == [999]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if if 999 else 321 then else 123 then ; 0 1 foo"
    )
    vm32.run()
    assert vm32.stack == [321]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if if 999 else 321 then else 123 then ; 1 1 foo"
    )
    vm32.run()
    assert vm32.stack == [999]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if 999 else if 123 else 321 then then ; -1 -1 foo"
    )
    vm32.run()
    assert vm32.stack == [-1, 999]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if 999 else if 123 else 321 then then ; 0 -1 foo"
    )
    vm32.run()
    assert vm32.stack == [0, 999]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if 999 else if 123 else 321 then then ; 1 -1 foo"
    )
    vm32.run()
    assert vm32.stack == [1, 999]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if 999 else if 123 else 321 then then ; -1 0 foo"
    )
    vm32.run()
    assert vm32.stack == [123]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if 999 else if 123 else 321 then then ; 0 0 foo"
    )
    vm32.run()
    assert vm32.stack == [321]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if 999 else if 123 else 321 then then ; 1 0 foo"
    )
    vm32.run()
    assert vm32.stack == [123]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if 999 else if 123 else 321 then then ; -1 1 foo"
    )
    vm32.run()
    assert vm32.stack == [-1, 999]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if 999 else if 123 else 321 then then ; 0 1 foo"
    )
    vm32.run()
    assert vm32.stack == [0, 999]

    vm32 = awkward.forth.ForthMachine32(
        ": foo if 999 else if 123 else 321 then then ; 1 1 foo"
    )
    vm32.run()
    assert vm32.stack == [1, 999]

    vm32 = awkward.forth.ForthMachine32(": foo do i loop ; 10 5 foo")
    vm32.run()
    assert vm32.stack == [5, 6, 7, 8, 9]

    vm32 = awkward.forth.ForthMachine32(": foo do i i +loop ; 100 5 foo")
    vm32.run()
    assert vm32.stack == [5, 10, 20, 40, 80]

    vm32 = awkward.forth.ForthMachine32(": foo 10 5 do 3 0 do 1+ loop loop ; 1 foo")
    vm32.run()
    assert vm32.stack == [16]

    vm32 = awkward.forth.ForthMachine32(": foo 10 5 do 3 0 do i loop loop ; foo")
    vm32.run()
    assert vm32.stack == [0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2]

    vm32 = awkward.forth.ForthMachine32(": foo 10 5 do 3 0 do j loop loop ; foo")
    vm32.run()
    assert vm32.stack == [5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9]

    vm32 = awkward.forth.ForthMachine32(": foo 10 5 do 3 0 do i j * loop loop ; foo")
    vm32.run()
    assert vm32.stack == [0, 5, 10, 0, 6, 12, 0, 7, 14, 0, 8, 16, 0, 9, 18]

    vm32 = awkward.forth.ForthMachine32(
        ": foo 10 5 do 8 6 do 3 0 do i j * k * loop loop loop ; foo"
    )
    vm32.run()
    assert vm32.stack == [
        0,
        30,
        60,
        0,
        35,
        70,
        0,
        36,
        72,
        0,
        42,
        84,
        0,
        42,
        84,
        0,
        49,
        98,
        0,
        48,
        96,
        0,
        56,
        112,
        0,
        54,
        108,
        0,
        63,
        126,
    ]

    vm32 = awkward.forth.ForthMachine32(": foo 3 begin dup 1 - dup 0= until ; foo")
    vm32.run()
    assert vm32.stack == [3, 2, 1, 0]

    vm32 = awkward.forth.ForthMachine32(
        ": foo 4 begin dup 1 - dup 0= invert while 123 drop repeat ; foo"
    )
    vm32.run()
    assert vm32.stack == [4, 3, 2, 1, 0]

    vm32 = awkward.forth.ForthMachine32(
        ": foo 3 begin dup 1 - dup 0= if exit then again ; foo"
    )
    vm32.run()
    assert vm32.stack == [3, 2, 1, 0]

    vm32 = awkward.forth.ForthMachine32("1 2 3 4 dup")
    vm32.run()
    assert vm32.stack == [1, 2, 3, 4, 4]

    vm32 = awkward.forth.ForthMachine32("1 2 3 4 drop")
    vm32.run()
    assert vm32.stack == [1, 2, 3]

    vm32 = awkward.forth.ForthMachine32("1 2 3 4 swap")
    vm32.run()
    assert vm32.stack == [1, 2, 4, 3]

    vm32 = awkward.forth.ForthMachine32("1 2 3 4 over")
    vm32.run()
    assert vm32.stack == [1, 2, 3, 4, 3]

    vm32 = awkward.forth.ForthMachine32("1 2 3 4 rot")
    vm32.run()
    assert vm32.stack == [1, 3, 4, 2]

    vm32 = awkward.forth.ForthMachine32("1 2 3 4 nip")
    vm32.run()
    assert vm32.stack == [1, 2, 4]

    vm32 = awkward.forth.ForthMachine32("1 2 3 4 tuck")
    vm32.run()
    assert vm32.stack == [1, 2, 4, 3, 4]

    vm32 = awkward.forth.ForthMachine32("3 5 +")
    vm32.run()
    assert vm32.stack == [8]

    vm32 = awkward.forth.ForthMachine32("-3 5 +")
    vm32.run()
    assert vm32.stack == [2]

    vm32 = awkward.forth.ForthMachine32("3 -5 +")
    vm32.run()
    assert vm32.stack == [-2]

    vm32 = awkward.forth.ForthMachine32("-3 -5 +")
    vm32.run()
    assert vm32.stack == [-8]

    vm32 = awkward.forth.ForthMachine32("3 5 -")
    vm32.run()
    assert vm32.stack == [-2]

    vm32 = awkward.forth.ForthMachine32("-3 5 -")
    vm32.run()
    assert vm32.stack == [-8]

    vm32 = awkward.forth.ForthMachine32("3 -5 -")
    vm32.run()
    assert vm32.stack == [8]

    vm32 = awkward.forth.ForthMachine32("-3 -5 -")
    vm32.run()
    assert vm32.stack == [2]

    vm32 = awkward.forth.ForthMachine32("5 3 -")
    vm32.run()
    assert vm32.stack == [2]

    vm32 = awkward.forth.ForthMachine32("5 -3 -")
    vm32.run()
    assert vm32.stack == [8]

    vm32 = awkward.forth.ForthMachine32("-5 3 -")
    vm32.run()
    assert vm32.stack == [-8]

    vm32 = awkward.forth.ForthMachine32("-5 -3 -")
    vm32.run()
    assert vm32.stack == [-2]

    vm32 = awkward.forth.ForthMachine32("3 5 *")
    vm32.run()
    assert vm32.stack == [15]

    vm32 = awkward.forth.ForthMachine32("-3 5 *")
    vm32.run()
    assert vm32.stack == [-15]

    vm32 = awkward.forth.ForthMachine32("3 -5 *")
    vm32.run()
    assert vm32.stack == [-15]

    vm32 = awkward.forth.ForthMachine32("-3 -5 *")
    vm32.run()
    assert vm32.stack == [15]

    vm32 = awkward.forth.ForthMachine32("22 7 /")
    vm32.run()
    assert vm32.stack == [3]

    vm32 = awkward.forth.ForthMachine32("-22 7 /")
    vm32.run()
    assert vm32.stack == [-4]

    vm32 = awkward.forth.ForthMachine32("22 -7 /")
    vm32.run()
    assert vm32.stack == [-4]

    vm32 = awkward.forth.ForthMachine32("-22 -7 /")
    vm32.run()
    assert vm32.stack == [3]

    vm32 = awkward.forth.ForthMachine32("22 7 mod")
    vm32.run()
    assert vm32.stack == [1]

    vm32 = awkward.forth.ForthMachine32("-22 7 mod")
    vm32.run()
    assert vm32.stack == [6]

    vm32 = awkward.forth.ForthMachine32("22 -7 mod")
    vm32.run()
    assert vm32.stack == [-6]

    vm32 = awkward.forth.ForthMachine32("-22 -7 mod")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("22 7 /mod")
    vm32.run()
    assert vm32.stack == [1, 3]

    vm32 = awkward.forth.ForthMachine32("-22 7 /mod")
    vm32.run()
    assert vm32.stack == [6, -4]

    vm32 = awkward.forth.ForthMachine32("22 -7 /mod")
    vm32.run()
    assert vm32.stack == [-6, -4]

    vm32 = awkward.forth.ForthMachine32("-22 -7 /mod")
    vm32.run()
    assert vm32.stack == [-1, 3]

    vm32 = awkward.forth.ForthMachine32("-2 abs")
    vm32.run()
    assert vm32.stack == [2]

    vm32 = awkward.forth.ForthMachine32("-1 abs")
    vm32.run()
    assert vm32.stack == [1]

    vm32 = awkward.forth.ForthMachine32("0 abs")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("1 abs")
    vm32.run()
    assert vm32.stack == [1]

    vm32 = awkward.forth.ForthMachine32("2 abs")
    vm32.run()
    assert vm32.stack == [2]

    vm32 = awkward.forth.ForthMachine32("5 5 min")
    vm32.run()
    assert vm32.stack == [5]

    vm32 = awkward.forth.ForthMachine32("3 -5 min")
    vm32.run()
    assert vm32.stack == [-5]

    vm32 = awkward.forth.ForthMachine32("-3 5 min")
    vm32.run()
    assert vm32.stack == [-3]

    vm32 = awkward.forth.ForthMachine32("3 5 min")
    vm32.run()
    assert vm32.stack == [3]

    vm32 = awkward.forth.ForthMachine32("5 5 max")
    vm32.run()
    assert vm32.stack == [5]

    vm32 = awkward.forth.ForthMachine32("3 -5 max")
    vm32.run()
    assert vm32.stack == [3]

    vm32 = awkward.forth.ForthMachine32("-3 5 max")
    vm32.run()
    assert vm32.stack == [5]

    vm32 = awkward.forth.ForthMachine32("3 5 max")
    vm32.run()
    assert vm32.stack == [5]

    vm32 = awkward.forth.ForthMachine32("-2 negate")
    vm32.run()
    assert vm32.stack == [2]

    vm32 = awkward.forth.ForthMachine32("-1 negate")
    vm32.run()
    assert vm32.stack == [1]

    vm32 = awkward.forth.ForthMachine32("0 negate")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("1 negate")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("2 negate")
    vm32.run()
    assert vm32.stack == [-2]

    vm32 = awkward.forth.ForthMachine32("-1 1+")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("0 1+")
    vm32.run()
    assert vm32.stack == [1]

    vm32 = awkward.forth.ForthMachine32("1 1+")
    vm32.run()
    assert vm32.stack == [2]

    vm32 = awkward.forth.ForthMachine32("-1 1-")
    vm32.run()
    assert vm32.stack == [-2]

    vm32 = awkward.forth.ForthMachine32("0 1-")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("1 1-")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("-1 0=")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("0 0=")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("1 0=")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("5 5 =")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("3 -5 =")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("-3 5 =")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("3 5 =")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("5 5 <>")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("3 -5 <>")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("-3 5 <>")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("3 5 <>")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("5 5 >")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("3 -5 >")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("-3 5 >")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("3 5 >")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("5 5 >=")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("3 -5 >=")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("-3 5 >=")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("3 5 >=")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("5 5 <")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("3 -5 <")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("-3 5 <")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("3 5 <")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("5 5 <=")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("3 -5 <=")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("-3 5 <=")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("3 5 <=")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("-1 -1 and")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("0 -1 and")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("1 -1 and")
    vm32.run()
    assert vm32.stack == [1]

    vm32 = awkward.forth.ForthMachine32("-1 0 and")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("0 0 and")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("1 0 and")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("-1 1 and")
    vm32.run()
    assert vm32.stack == [1]

    vm32 = awkward.forth.ForthMachine32("0 1 and")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("1 1 and")
    vm32.run()
    assert vm32.stack == [1]

    vm32 = awkward.forth.ForthMachine32("-1 -1 or")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("0 -1 or")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("1 -1 or")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("-1 0 or")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("0 0 or")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("1 0 or")
    vm32.run()
    assert vm32.stack == [1]

    vm32 = awkward.forth.ForthMachine32("-1 1 or")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("0 1 or")
    vm32.run()
    assert vm32.stack == [1]

    vm32 = awkward.forth.ForthMachine32("1 1 or")
    vm32.run()
    assert vm32.stack == [1]

    vm32 = awkward.forth.ForthMachine32("-1 -1 xor")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("0 -1 xor")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("1 -1 xor")
    vm32.run()
    assert vm32.stack == [-2]

    vm32 = awkward.forth.ForthMachine32("-1 0 xor")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("0 0 xor")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("1 0 xor")
    vm32.run()
    assert vm32.stack == [1]

    vm32 = awkward.forth.ForthMachine32("-1 1 xor")
    vm32.run()
    assert vm32.stack == [-2]

    vm32 = awkward.forth.ForthMachine32("0 1 xor")
    vm32.run()
    assert vm32.stack == [1]

    vm32 = awkward.forth.ForthMachine32("1 1 xor")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("-1 invert")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("0 invert")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("1 invert")
    vm32.run()
    assert vm32.stack == [-2]

    vm32 = awkward.forth.ForthMachine32("true")
    vm32.run()
    assert vm32.stack == [-1]

    vm32 = awkward.forth.ForthMachine32("false")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32(": foo 3 + 2 * ; 4 foo foo")
    vm32.run()
    assert vm32.stack == [34]

    vm32 = awkward.forth.ForthMachine32(": foo 3 + 2 * ; : bar 4 foo foo ; bar")
    vm32.run()
    assert vm32.stack == [34]

    vm32 = awkward.forth.ForthMachine32(
        ": factorial dup 2 < if drop 1 exit then dup 1- recurse * ; 5 factorial"
    )
    vm32.run()
    assert vm32.stack == [120]

    vm32 = awkward.forth.ForthMachine32("variable x 10 x ! 5 x +! x @ x @ x @")
    vm32.run()
    assert vm32.stack == [15, 15, 15]

    vm32 = awkward.forth.ForthMachine32(
        "variable x 10 x ! 5 x +! : foo x @ x @ x @ ; foo foo"
    )
    vm32.run()
    assert vm32.stack == [15, 15, 15, 15, 15, 15]


def test_gforth_lshift_rshift():
    vm32 = awkward.forth.ForthMachine32("0 1 lshift")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("0 2 lshift")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("0 3 lshift")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("1 1 lshift")
    vm32.run()
    assert vm32.stack == [2]

    vm32 = awkward.forth.ForthMachine32("1 2 lshift")
    vm32.run()
    assert vm32.stack == [4]

    vm32 = awkward.forth.ForthMachine32("1 3 lshift")
    vm32.run()
    assert vm32.stack == [8]

    vm32 = awkward.forth.ForthMachine32("5 1 lshift")
    vm32.run()
    assert vm32.stack == [10]

    vm32 = awkward.forth.ForthMachine32("5 2 lshift")
    vm32.run()
    assert vm32.stack == [20]

    vm32 = awkward.forth.ForthMachine32("5 3 lshift")
    vm32.run()
    assert vm32.stack == [40]

    vm32 = awkward.forth.ForthMachine32("-5 1 lshift")
    vm32.run()
    assert vm32.stack == [-10]

    vm32 = awkward.forth.ForthMachine32("-5 2 lshift")
    vm32.run()
    assert vm32.stack == [-20]

    vm32 = awkward.forth.ForthMachine32("-5 3 lshift")
    vm32.run()
    assert vm32.stack == [-40]

    vm32 = awkward.forth.ForthMachine32("0 1 rshift")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("0 2 rshift")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("0 3 rshift")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("1 1 rshift")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("1 2 rshift")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("1 3 rshift")
    vm32.run()
    assert vm32.stack == [0]

    vm32 = awkward.forth.ForthMachine32("5 1 rshift")
    vm32.run()
    assert vm32.stack == [2]

    vm32 = awkward.forth.ForthMachine32("5 2 rshift")
    vm32.run()
    assert vm32.stack == [1]

    vm32 = awkward.forth.ForthMachine32("5 3 rshift")
    vm32.run()
    assert vm32.stack == [0]

    # I don't understand what gforth's rshift is doing with negative numbers.
    # These can't be brought into agreement even with 64-bit machines.
    # Maybe gforth is just wrong: I would expect the results WE get, not THEIRS.

    vm32 = awkward.forth.ForthMachine32("-5 1 rshift")
    vm32.run()
    # assert vm32.stack == [9223372036854775805]
    assert vm32.stack == [-3]

    vm32 = awkward.forth.ForthMachine32("-5 2 rshift")
    vm32.run()
    # assert vm32.stack == [4611686018427387902]
    assert vm32.stack == [-2]

    vm32 = awkward.forth.ForthMachine32("-5 3 rshift")
    vm32.run()
    # assert vm32.stack == [2305843009213693951]
    assert vm32.stack == [-1]


def test_input_output():
    vm32 = awkward.forth.ForthMachine32(
        """
output y float32
4 3 2 1
y <- stack
y <- stack
y <- stack
y <- stack
"""
    )
    vm32.run()
    assert vm32.stack == []
    assert ak.to_list(vm32["y"]) == [1, 2, 3, 4]

    vm32 = awkward.forth.ForthMachine32(
        """
output y float32
4 3 2 1
y <- stack
y <- stack
1 y rewind
y <- stack
y <- stack
"""
    )
    vm32.run()
    assert ak.to_list(vm32["y"]) == [1, 3, 4]

    vm32 = awkward.forth.ForthMachine32(
        """
input x
x i-> stack
x i-> stack
-8 x skip
x i-> stack
x i-> stack
x i-> stack
x i-> stack
"""
    )
    vm32.run({"x": np.array([1, 2, 3, 4], np.int32)})
    assert vm32.stack == [1, 2, 1, 2, 3, 4]

    vm32 = awkward.forth.ForthMachine32(
        """
input x
4 x #i-> stack
"""
    )
    vm32.run({"x": np.array([1, 2, 3, 4], np.int32)})
    assert vm32.stack == [1, 2, 3, 4]

    vm32 = awkward.forth.ForthMachine32(
        """
input x
4 x #!i-> stack
"""
    )
    vm32.run({"x": np.array([1, 2, 3, 4], np.int32)})
    assert vm32.stack == [16777216, 33554432, 50331648, 67108864]

    vm32 = awkward.forth.ForthMachine32(
        """
input x
8 x #h-> stack
"""
    )
    vm32.run({"x": np.array([1, 2, 3, 4], np.int32)})
    assert vm32.stack == [1, 0, 2, 0, 3, 0, 4, 0]

    vm32 = awkward.forth.ForthMachine32(
        """
input x output y float32
x i-> y
x i-> y
x i-> y
x i-> y
"""
    )
    vm32.run({"x": np.array([1, 2, 3, 4], np.int32)})
    assert vm32.stack == []
    assert ak.to_list(vm32["y"]) == [1, 2, 3, 4]

    vm32 = awkward.forth.ForthMachine32(
        """
input x output y float32
4 x #i-> y
"""
    )
    vm32.run({"x": np.array([1, 2, 3, 4], np.int32)})
    assert vm32.stack == []
    assert ak.to_list(vm32["y"]) == [1, 2, 3, 4]


def test_uproot():
    uproot = pytest.importorskip("uproot")
    skhep_testdata = pytest.importorskip("skhep_testdata")

    vector_of_vectors = uproot.open(
        skhep_testdata.data_path("uproot-vectorVectorDouble.root")
    )["t/x"]
    inputs = {
        "data": np.array(vector_of_vectors.basket(0).data),
        "byte_offsets": vector_of_vectors.basket(0).byte_offsets,
    }

    vm32 = awkward.forth.ForthMachine32(
        """
input data
input byte_offsets
output outer-offsets int32
output inner-offsets int32
output content-data float64
variable length
variable outer-last
variable inner-last

0 length !
0 outer-last !
0 inner-last !
outer-last @ outer-offsets <- stack
inner-last @ inner-offsets <- stack

begin
  byte_offsets i-> stack
  6 + data seek
  1 length +!

  data !i-> stack dup outer-last +!
  outer-last @ outer-offsets <- stack

  0 do
    data !i-> stack dup inner-last +!
    inner-last @ inner-offsets <- stack

    data #!d-> content-data
  loop
again
"""
    )
    vm32.run(inputs, raise_read_beyond=False, raise_seek_beyond=False)
    assert ak.to_list(vm32["outer-offsets"]) == [0, 0, 2, 5, 6, 9]
    assert ak.to_list(vm32["inner-offsets"]) == [0, 0, 0, 1, 1, 3, 6, 7, 8, 9]
    assert ak.to_list(vm32["content-data"]) == [
        10,
        10,
        20,
        20,
        -21,
        -22,
        200,
        -201,
        202,
    ]


def test_varint():
    # https://lucene.apache.org/core/3_5_0/fileformats.html#VInt
    data = np.array(
        [0, 1, 2, 127, 128, 1, 129, 1, 130, 1, 255, 127, 128, 128, 1, 129, 128, 1],
        np.uint8,
    )

    vm32 = awkward.forth.ForthMachine32(
        """
input x
10 0 do
  x varint-> stack
loop
"""
    )
    vm32.run({"x": data})
    assert vm32.stack == [0, 1, 2, 127, 128, 129, 130, 16383, 16384, 16385]

    vm32 = awkward.forth.ForthMachine32(
        """
input x
10 x #varint-> stack
"""
    )
    vm32.run({"x": data})
    assert vm32.stack == [0, 1, 2, 127, 128, 129, 130, 16383, 16384, 16385]

    vm32 = awkward.forth.ForthMachine32(
        """
input x
output y int32
10 0 do
  x varint-> y
loop
"""
    )
    vm32.run({"x": data})
    assert ak.to_list(vm32["y"]) == [0, 1, 2, 127, 128, 129, 130, 16383, 16384, 16385]

    vm32 = awkward.forth.ForthMachine32(
        """
input x
output y int32
10 x #varint-> y
"""
    )
    vm32.run({"x": data})
    assert ak.to_list(vm32["y"]) == [0, 1, 2, 127, 128, 129, 130, 16383, 16384, 16385]


def test_zigzag():
    # https://developers.google.com/protocol-buffers/docs/encoding?csw=1#types
    data = np.array(
        [0, 1, 2, 3, 254, 255, 255, 255, 15, 255, 255, 255, 255, 15], np.uint8
    )

    vm32 = awkward.forth.ForthMachine32(
        """
input x
6 0 do
  x zigzag-> stack
loop
"""
    )
    vm32.run({"x": data})
    assert vm32.stack == [0, -1, 1, -2, 2147483647, -2147483648]

    vm32 = awkward.forth.ForthMachine32(
        """
input x
6 x #zigzag-> stack
"""
    )
    vm32.run({"x": data})
    assert vm32.stack == [0, -1, 1, -2, 2147483647, -2147483648]

    vm32 = awkward.forth.ForthMachine32(
        """
input x
output y int32
6 0 do
  x zigzag-> y
loop
"""
    )
    vm32.run({"x": data})
    assert ak.to_list(vm32["y"]) == [0, -1, 1, -2, 2147483647, -2147483648]

    vm32 = awkward.forth.ForthMachine32(
        """
input x
output y int32
6 x #zigzag-> y
"""
    )
    vm32.run({"x": data})
    assert ak.to_list(vm32["y"]) == [0, -1, 1, -2, 2147483647, -2147483648]


def test_nbit_little():
    data = np.array(
        [252, 255, 191, 255, 251, 255, 255, 239, 251, 255, 191, 255], np.uint8
    )
    expectation = (
        [0, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2]
        + [3, 3, 3, 3, 3, 2, 3, 3, 3, 3, 3, 3]
        + [3, 3, 3, 3, 3, 3, 2, 3, 3, 2, 3, 3]
        + [3, 3, 3, 3, 3, 3, 3, 2, 3, 3, 3, 3]
    )

    vm32 = awkward.forth.ForthMachine32("input x 48 x #2bit-> stack")
    vm32.run({"x": data})
    assert vm32.stack == expectation
    assert vm32.input_position("x") == 12

    vm32 = awkward.forth.ForthMachine32("input x 47 x #2bit-> stack")
    vm32.run({"x": data})
    assert vm32.stack == expectation[:47]
    assert vm32.input_position("x") == 12

    vm32 = awkward.forth.ForthMachine32("input x 45 x #2bit-> stack")
    vm32.run({"x": data})
    assert vm32.stack == expectation[:45]
    assert vm32.input_position("x") == 12

    vm32 = awkward.forth.ForthMachine32("input x 44 x #2bit-> stack")
    vm32.run({"x": data})
    assert vm32.stack == expectation[:44]
    assert vm32.input_position("x") == 11

    vm32 = awkward.forth.ForthMachine32("input x 43 x #2bit-> stack")
    vm32.run({"x": data})
    assert vm32.stack == expectation[:43]
    assert vm32.input_position("x") == 11

    vm32 = awkward.forth.ForthMachine32("input x 1 x #2bit-> stack")
    vm32.run({"x": data})
    assert vm32.stack == expectation[:1]
    assert vm32.input_position("x") == 1

    vm32 = awkward.forth.ForthMachine32("input x x 2bit-> stack")
    vm32.run({"x": data})
    assert vm32.stack == expectation[:1]
    assert vm32.input_position("x") == 1

    vm32 = awkward.forth.ForthMachine32("input x 0 x #2bit-> stack")
    vm32.run({"x": data})
    assert vm32.stack == []
    assert vm32.input_position("x") == 0

    vm32 = awkward.forth.ForthMachine32("input x output y int32 48 x #2bit-> y")
    vm32.run({"x": data})
    assert ak.to_list(vm32["y"]) == expectation
    assert vm32.input_position("x") == 12

    vm32 = awkward.forth.ForthMachine32("input x output y int32 x 2bit-> y")
    vm32.run({"x": data})
    assert ak.to_list(vm32["y"]) == expectation[:1]
    assert vm32.input_position("x") == 1


def test_nbit_big():
    data = np.array(
        [63, 255, 253, 255, 223, 255, 255, 247, 223, 255, 253, 255], np.uint8
    )
    expectation = (
        [0, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2]
        + [3, 3, 3, 3, 3, 2, 3, 3, 3, 3, 3, 3]
        + [3, 3, 3, 3, 3, 3, 2, 3, 3, 2, 3, 3]
        + [3, 3, 3, 3, 3, 3, 3, 2, 3, 3, 3, 3]
    )

    vm32 = awkward.forth.ForthMachine32("input x 48 x #!2bit-> stack")
    vm32.run({"x": data})
    assert vm32.stack == expectation
    assert vm32.input_position("x") == 12

    vm32 = awkward.forth.ForthMachine32("input x 47 x #!2bit-> stack")
    vm32.run({"x": data})
    assert vm32.stack == expectation[:47]
    assert vm32.input_position("x") == 12

    vm32 = awkward.forth.ForthMachine32("input x 45 x #!2bit-> stack")
    vm32.run({"x": data})
    assert vm32.stack == expectation[:45]
    assert vm32.input_position("x") == 12

    vm32 = awkward.forth.ForthMachine32("input x 44 x #!2bit-> stack")
    vm32.run({"x": data})
    assert vm32.stack == expectation[:44]
    assert vm32.input_position("x") == 11

    vm32 = awkward.forth.ForthMachine32("input x 43 x #!2bit-> stack")
    vm32.run({"x": data})
    assert vm32.stack == expectation[:43]
    assert vm32.input_position("x") == 11

    vm32 = awkward.forth.ForthMachine32("input x 1 x #!2bit-> stack")
    vm32.run({"x": data})
    assert vm32.stack == expectation[:1]
    assert vm32.input_position("x") == 1

    vm32 = awkward.forth.ForthMachine32("input x x !2bit-> stack")
    vm32.run({"x": data})
    assert vm32.stack == expectation[:1]
    assert vm32.input_position("x") == 1

    vm32 = awkward.forth.ForthMachine32("input x 0 x #!2bit-> stack")
    vm32.run({"x": data})
    assert vm32.stack == []
    assert vm32.input_position("x") == 0

    vm32 = awkward.forth.ForthMachine32("input x output y int32 48 x #!2bit-> y")
    vm32.run({"x": data})
    assert ak.to_list(vm32["y"]) == expectation
    assert vm32.input_position("x") == 12

    vm32 = awkward.forth.ForthMachine32("input x output y int32 x !2bit-> y")
    vm32.run({"x": data})
    assert ak.to_list(vm32["y"]) == expectation[:1]
    assert vm32.input_position("x") == 1


def test_output_dup():
    vm = awkward.forth.ForthMachine32(
        """
output stuff int32
1 2 3 4
stuff <- stack
10 stuff dup
"""
    )
    vm.run()
    assert ak.to_list(vm["stuff"]) == [4] * 11
    assert vm.stack == [1, 2, 3]


def test_decompile_complex():
    read_repetition_levels = awkward.forth.ForthMachine32(
        """
input stream
output replevels uint8

stream I-> stack
begin
  stream varint-> stack
  dup 1 and 0= if
    ( run-length encoding )
    stream {rle_format}-> replevels
    1 rshift 1-
    replevels dup
  else
    ( bit-packed )
    1 rshift 8 *
    stream #{bit_width}bit-> replevels
  then
  dup stream pos 4 - <=
until
""".format(
            bit_width=2, rle_format="B"
        )
    )
    assert (
        read_repetition_levels.decompiled
        == """input stream
output replevels uint8

stream I-> stack
begin
  stream varint-> stack
  dup
  1
  and
  0=
  if
    stream B-> replevels
    1
    rshift
    1-
    replevels dup
  else
    1
    rshift
    8
    *
    stream #2bit-> replevels
  then
  dup
  stream pos
  4
  -
  <=
until
"""
    )


def test_parse_strings():
    vm32 = awkward.forth.ForthMachine32(
        """
." hello there      long space"
"""
    )
    assert (
        vm32.decompiled
        == """." hello there      long space"
"""
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


class ByteBehavior(ak.highlevel.Array):
    __name__ = "Array"

    def __bytes__(self):
        tmp = ak.nplike.of(self.layout).asarray(self.layout)
        if hasattr(tmp, "tobytes"):
            return tmp.tobytes()
        else:
            return tmp.tostring()

    def __str__(self):
        return str(self.__bytes__())

    def __repr__(self):
        return repr(self.__bytes__())

    def __iter__(self):
        for x in self.__bytes__():
            yield x

    def __eq__(self, other):
        if isinstance(other, (bytes, ByteBehavior)):
            return bytes(self) == bytes(other)
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(self, other)

    def __add__(self, other):
        if isinstance(other, (bytes, ByteBehavior)):
            return bytes(self) + bytes(other)
        else:
            raise TypeError("can only concatenate bytes to bytes")

    def __radd__(self, other):
        if isinstance(other, (bytes, ByteBehavior)):
            return bytes(other) + bytes(self)
        else:
            raise TypeError("can only concatenate bytes to bytes")


class CharBehavior(ak.highlevel.Array):
    __name__ = "Array"

    def __bytes__(self):
        tmp = ak.nplike.of(self.layout).asarray(self.layout)
        if hasattr(tmp, "tobytes"):
            return tmp.tobytes()
        else:
            return tmp.tostring()

    def __str__(self):
        return self.__bytes__().decode("utf-8", "surrogateescape")

    def __repr__(self):
        return repr(self.__bytes__().decode("utf-8", "surrogateescape"))

    def __iter__(self):
        for x in self.__str__():
            yield x

    def __eq__(self, other):
        if isinstance(other, (str, CharBehavior)):
            return str(self) == str(other)
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __add__(self, other):
        if isinstance(other, (str, CharBehavior)):
            return str(self) + str(other)
        else:
            raise TypeError("can only concatenate str to str")

    def __radd__(self, other):
        if isinstance(other, (str, CharBehavior)):
            return str(other) + str(self)
        else:
            raise TypeError("can only concatenate str to str")


ak.behavior["byte"] = ByteBehavior
ak.behavior["__typestr__", "byte"] = "byte"
ak.behavior["char"] = CharBehavior
ak.behavior["__typestr__", "char"] = "char"


class ByteStringBehavior(ak.highlevel.Array):
    __name__ = "Array"

    def __iter__(self):
        for x in super(ByteStringBehavior, self).__iter__():
            yield x.__bytes__()


class StringBehavior(ak.highlevel.Array):
    __name__ = "Array"

    def __iter__(self):
        for x in super(StringBehavior, self).__iter__():
            yield x.__str__()


ak.behavior["bytestring"] = ByteStringBehavior
ak.behavior["__typestr__", "bytestring"] = "bytes"
ak.behavior["string"] = StringBehavior
ak.behavior["__typestr__", "string"] = "string"


def _string_equal(one, two):
    nplike = ak.nplike.of(one, two)
    behavior = ak._util.behaviorof(one, two)

    one, two = ak.without_parameters(one).layout, ak.without_parameters(two).layout

    # first condition: string lengths must be the same
    counts1 = nplike.asarray(one.count(axis=-1))
    counts2 = nplike.asarray(two.count(axis=-1))

    out = counts1 == counts2

    # only compare characters in strings that are possibly equal (same length)
    possible = nplike.logical_and(out, counts1)
    possible_counts = counts1[possible]

    if len(possible_counts) > 0:
        onepossible = one[possible]
        twopossible = two[possible]

        reduced = ak.all(ak.Array(onepossible) == ak.Array(twopossible), axis=-1).layout

        # update same-length strings with a verdict about their characters
        out[possible] = reduced

    return ak._util.wrap(ak.layout.NumpyArray(out), behavior)


def _string_notequal(one, two):
    return ~_string_equal(one, two)


ak.behavior[ak.nplike.numpy.equal, "bytestring", "bytestring"] = _string_equal
ak.behavior[ak.nplike.numpy.equal, "string", "string"] = _string_equal
ak.behavior[ak.nplike.numpy.not_equal, "bytestring", "bytestring"] = _string_notequal
ak.behavior[ak.nplike.numpy.not_equal, "string", "string"] = _string_notequal


def _string_broadcast(layout, offsets):
    nplike = ak.nplike.of(offsets)
    offsets = nplike.asarray(offsets)
    counts = offsets[1:] - offsets[:-1]
    if ak._util.win or ak._util.bits32:
        counts = counts.astype(np.int32)
    parents = nplike.repeat(nplike.arange(len(counts), dtype=counts.dtype), counts)
    return ak.layout.IndexedArray64(ak.layout.Index64(parents), layout).project()


ak.behavior["__broadcast__", "bytestring"] = _string_broadcast
ak.behavior["__broadcast__", "string"] = _string_broadcast


def _string_numba_typer(viewtype):
    import numba

    if viewtype.type.parameters["__array__"] == "string":
        return numba.types.string
    else:
        return numba.cpython.charseq.bytes_type


def _string_numba_lower(
    context, builder, rettype, viewtype, viewval, viewproxy, attype, atval
):
    import numba
    import llvmlite.llvmpy.core

    whichpos = ak._connect._numba.layout.posat(
        context, builder, viewproxy.pos, viewtype.type.CONTENT
    )
    nextpos = ak._connect._numba.layout.getat(
        context, builder, viewproxy.arrayptrs, whichpos
    )

    whichnextpos = ak._connect._numba.layout.posat(
        context, builder, nextpos, viewtype.type.contenttype.ARRAY
    )

    startspos = ak._connect._numba.layout.posat(
        context, builder, viewproxy.pos, viewtype.type.STARTS
    )
    startsptr = ak._connect._numba.layout.getat(
        context, builder, viewproxy.arrayptrs, startspos
    )
    startsarraypos = builder.add(viewproxy.start, atval)
    start = ak._connect._numba.layout.getat(
        context, builder, startsptr, startsarraypos, viewtype.type.indextype.dtype
    )

    stopspos = ak._connect._numba.layout.posat(
        context, builder, viewproxy.pos, viewtype.type.STOPS
    )
    stopsptr = ak._connect._numba.layout.getat(
        context, builder, viewproxy.arrayptrs, stopspos
    )
    stopsarraypos = builder.add(viewproxy.start, atval)
    stop = ak._connect._numba.layout.getat(
        context, builder, stopsptr, stopsarraypos, viewtype.type.indextype.dtype
    )

    baseptr = ak._connect._numba.layout.getat(
        context, builder, viewproxy.arrayptrs, whichnextpos
    )
    rawptr = builder.add(
        baseptr,
        ak._connect._numba.castint(
            context, builder, viewtype.type.indextype.dtype, numba.intp, start
        ),
    )
    rawptr_cast = builder.inttoptr(
        rawptr,
        llvmlite.llvmpy.core.Type.pointer(
            llvmlite.llvmpy.core.Type.int(numba.intp.bitwidth // 8)
        ),
    )
    strsize = builder.sub(stop, start)
    strsize_cast = ak._connect._numba.castint(
        context, builder, viewtype.type.indextype.dtype, numba.intp, strsize
    )

    pyapi = context.get_python_api(builder)
    gil = pyapi.gil_ensure()

    strptr = builder.bitcast(rawptr_cast, pyapi.cstring)

    if viewtype.type.parameters["__array__"] == "string":
        kind = context.get_constant(numba.types.int32, pyapi.py_unicode_1byte_kind)
        pystr = pyapi.string_from_kind_and_data(kind, strptr, strsize_cast)
    else:
        pystr = pyapi.bytes_from_string_and_size(strptr, strsize_cast)

    out = pyapi.to_native_value(rettype, pystr).value

    pyapi.gil_release(gil)

    return out


ak.behavior["__numba_typer__", "bytestring"] = _string_numba_typer
ak.behavior["__numba_lower__", "bytestring"] = _string_numba_lower
ak.behavior["__numba_typer__", "string"] = _string_numba_typer
ak.behavior["__numba_lower__", "string"] = _string_numba_lower


__all__ = [
    x for x in list(globals()) if not x.startswith("_") and x not in ("ak", "np")
]


def __dir__():
    return __all__



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

from awkward._v2.forms.form import Form, from_iter, from_json  # noqa: F401
from awkward._v2.forms.emptyform import EmptyForm  # noqa: F401
from awkward._v2.forms.numpyform import NumpyForm, from_dtype  # noqa: F401
from awkward._v2.forms.regularform import RegularForm  # noqa: F401
from awkward._v2.forms.listform import ListForm  # noqa: F401
from awkward._v2.forms.listoffsetform import ListOffsetForm  # noqa: F401
from awkward._v2.forms.recordform import RecordForm  # noqa: F401
from awkward._v2.forms.indexedform import IndexedForm  # noqa: F401
from awkward._v2.forms.indexedoptionform import IndexedOptionForm  # noqa: F401
from awkward._v2.forms.bytemaskedform import ByteMaskedForm  # noqa: F401
from awkward._v2.forms.bitmaskedform import BitMaskedForm  # noqa: F401
from awkward._v2.forms.unmaskedform import UnmaskedForm  # noqa: F401
from awkward._v2.forms.unionform import UnionForm  # noqa: F401
from awkward._v2.forms.virtualform import VirtualForm  # noqa: F401



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


empty = ak.Array(
    ak.layout.RegularArray(ak.Array([[1, 2, 3], [], [4, 5]]).layout, 0, zeros_length=0)
)


def test_ListOffsetArray_rpad_and_clip():
    array = ak.Array([[1, 2, 3], [], [4, 5]])
    assert ak.pad_none(array, 0, clip=True).tolist() == [[], [], []]

    array = ak.Array([[1, 2, 3], [], [4, 5]])
    assert ak.pad_none(array, 0).tolist() == [[1, 2, 3], [], [4, 5]]


def test_toListOffsetArray64():
    assert ak.from_regular(empty).tolist() == []


def test_setidentities():
    empty2 = ak.Array(
        ak.layout.RegularArray(
            ak.Array([[1, 2, 3], [], [4, 5]]).layout, 0, zeros_length=0
        )
    )
    empty2.layout.setidentities()
    assert np.asarray(empty2.layout.identities).tolist() == []


def test_carry():
    assert empty[[]].tolist() == []


def test_num():
    assert ak.num(empty, axis=0) == 0
    assert ak.num(empty, axis=1).tolist() == []
    assert ak.num(empty, axis=2).tolist() == []


def test_flatten():
    assert ak.flatten(empty, axis=0).tolist() == []
    assert ak.flatten(empty, axis=1).tolist() == []
    assert ak.flatten(empty, axis=2).tolist() == []


def test_mergeable():
    assert ak.concatenate([empty, empty]).tolist() == []


def test_fillna():
    assert ak.fill_none(empty, 5, axis=0).tolist() == []


def test_pad_none():
    assert ak.pad_none(empty, 0, axis=0).tolist() == []
    assert ak.pad_none(empty, 0, axis=1).tolist() == []
    assert ak.pad_none(empty, 0, axis=2).tolist() == []

    assert ak.pad_none(empty, 1, axis=0).tolist() == [None]
    assert ak.pad_none(empty, 1, axis=1).tolist() == []
    assert ak.pad_none(empty, 1, axis=2).tolist() == []

    assert ak.pad_none(empty, 0, axis=0, clip=True).tolist() == []
    assert ak.pad_none(empty, 0, axis=1, clip=True).tolist() == []
    assert ak.pad_none(empty, 0, axis=2, clip=True).tolist() == []

    assert ak.pad_none(empty, 1, axis=0, clip=True).tolist() == [None]
    assert ak.pad_none(empty, 1, axis=1, clip=True).tolist() == []
    assert ak.pad_none(empty, 1, axis=2, clip=True).tolist() == []


def test_reduce():
    assert ak.sum(empty, axis=0).tolist() == []


def test_localindex():
    assert ak.local_index(empty, axis=0).tolist() == []
    assert ak.local_index(empty, axis=1).tolist() == []
    assert ak.local_index(empty, axis=2).tolist() == []


def test_combinations():
    assert ak.combinations(empty, 2, axis=0).tolist() == []
    assert ak.combinations(empty, 2, axis=1).tolist() == []
    assert ak.combinations(empty, 2, axis=2).tolist() == []


def test_getitem():
    with pytest.raises(ValueError):
        empty[
            0,
        ]

    jagged = ak.Array([[]])[0:0]
    assert empty[jagged].tolist() == []



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


# https://github.com/scikit-hep/awkward-1.0/issues/459#issuecomment-694941328
#
# So the rules would be,
#    * if arrays have different `__array__` or `__record__` parameters, they are not equal;
#    * if they otherwise have different parameters, the types can be equal, but merging
#      (concatenation, option-simplify, or union-simplify) removes parameters other than
#      `__array__` and `__record__`.


def test_0459_types():
    plain_plain = ak.Array([0.0, 1.1, 2.2, 3.3, 4.4])
    array_plain = ak.with_parameter(plain_plain, "__array__", "zoinks")
    plain_isdoc = ak.with_parameter(plain_plain, "__doc__", "This is a zoink.")
    array_isdoc = ak.with_parameter(array_plain, "__doc__", "This is a zoink.")
    assert ak.parameters(plain_plain) == {}
    assert ak.parameters(array_plain) == {"__array__": "zoinks"}
    assert ak.parameters(plain_isdoc) == {"__doc__": "This is a zoink."}
    assert ak.parameters(array_isdoc) == {
        "__array__": "zoinks",
        "__doc__": "This is a zoink.",
    }

    assert ak.type(plain_plain) == ak.type(plain_plain)
    assert ak.type(array_plain) == ak.type(array_plain)
    assert ak.type(plain_isdoc) == ak.type(plain_isdoc)
    assert ak.type(array_isdoc) == ak.type(array_isdoc)

    assert ak.type(plain_plain) != ak.type(array_plain)
    assert ak.type(array_plain) != ak.type(plain_plain)

    assert ak.type(plain_plain) == ak.type(plain_isdoc)
    assert ak.type(plain_isdoc) == ak.type(plain_plain)

    assert ak.type(array_plain) == ak.type(array_isdoc)
    assert ak.type(array_isdoc) == ak.type(array_plain)

    assert ak.type(plain_isdoc) != ak.type(array_isdoc)
    assert ak.type(array_isdoc) != ak.type(plain_isdoc)


def test_0459():
    plain_plain = ak.Array([0.0, 1.1, 2.2, 3.3, 4.4])
    array_plain = ak.with_parameter(plain_plain, "__array__", "zoinks")
    plain_isdoc = ak.with_parameter(plain_plain, "__doc__", "This is a zoink.")
    array_isdoc = ak.with_parameter(array_plain, "__doc__", "This is a zoink.")
    assert ak.parameters(plain_plain) == {}
    assert ak.parameters(array_plain) == {"__array__": "zoinks"}
    assert ak.parameters(plain_isdoc) == {"__doc__": "This is a zoink."}
    assert ak.parameters(array_isdoc) == {
        "__array__": "zoinks",
        "__doc__": "This is a zoink.",
    }

    assert ak.parameters(ak.concatenate([plain_plain, plain_plain])) == {}
    assert ak.parameters(ak.concatenate([array_plain, array_plain])) == {
        "__array__": "zoinks"
    }
    assert ak.parameters(ak.concatenate([plain_isdoc, plain_isdoc])) == {
        "__doc__": "This is a zoink."
    }
    assert ak.parameters(ak.concatenate([array_isdoc, array_isdoc])) == {
        "__array__": "zoinks",
        "__doc__": "This is a zoink.",
    }

    assert isinstance(
        ak.concatenate([plain_plain, plain_plain]).layout, ak.layout.NumpyArray
    )
    assert isinstance(
        ak.concatenate([array_plain, array_plain]).layout, ak.layout.NumpyArray
    )
    assert isinstance(
        ak.concatenate([plain_isdoc, plain_isdoc]).layout, ak.layout.NumpyArray
    )
    assert isinstance(
        ak.concatenate([array_isdoc, array_isdoc]).layout, ak.layout.NumpyArray
    )

    assert ak.parameters(ak.concatenate([plain_plain, array_plain])) == {}
    assert ak.parameters(ak.concatenate([plain_isdoc, array_isdoc])) == {}
    assert ak.parameters(ak.concatenate([array_plain, plain_plain])) == {}
    assert ak.parameters(ak.concatenate([array_isdoc, plain_isdoc])) == {}

    assert isinstance(
        ak.concatenate([plain_plain, array_plain]).layout, ak.layout.UnionArray8_64
    )
    assert isinstance(
        ak.concatenate([plain_isdoc, array_isdoc]).layout, ak.layout.UnionArray8_64
    )
    assert isinstance(
        ak.concatenate([array_plain, plain_plain]).layout, ak.layout.UnionArray8_64
    )
    assert isinstance(
        ak.concatenate([array_isdoc, plain_isdoc]).layout, ak.layout.UnionArray8_64
    )

    assert ak.parameters(ak.concatenate([plain_plain, plain_isdoc])) == {}
    assert ak.parameters(ak.concatenate([array_plain, array_isdoc])) == {
        "__array__": "zoinks"
    }
    assert ak.parameters(ak.concatenate([plain_isdoc, plain_plain])) == {}
    assert ak.parameters(ak.concatenate([array_isdoc, array_plain])) == {
        "__array__": "zoinks"
    }

    assert isinstance(
        ak.concatenate([plain_plain, plain_isdoc]).layout, ak.layout.NumpyArray
    )
    assert isinstance(
        ak.concatenate([array_plain, array_isdoc]).layout, ak.layout.NumpyArray
    )
    assert isinstance(
        ak.concatenate([plain_isdoc, plain_plain]).layout, ak.layout.NumpyArray
    )
    assert isinstance(
        ak.concatenate([array_isdoc, array_plain]).layout, ak.layout.NumpyArray
    )


def test_0522():
    content1 = ak.Array([0.0, 1.1, 2.2, 3.3, 4.4]).layout
    content2 = ak.Array([0, 100, 200, 300, 400]).layout
    tags = ak.layout.Index8(np.array([0, 0, 0, 1, 1, 0, 0, 1, 1, 1], np.int8))
    index = ak.layout.Index64(np.array([0, 1, 2, 0, 1, 3, 4, 2, 3, 4], np.int64))
    unionarray = ak.Array(ak.layout.UnionArray8_64(tags, index, [content1, content2]))
    assert unionarray.tolist() == [0.0, 1.1, 2.2, 0, 100, 3.3, 4.4, 200, 300, 400]

    assert (unionarray + 10).tolist() == [
        10.0,
        11.1,
        12.2,
        10,
        110,
        13.3,
        14.4,
        210,
        310,
        410,
    ]
    assert (10 + unionarray).tolist() == [
        10.0,
        11.1,
        12.2,
        10,
        110,
        13.3,
        14.4,
        210,
        310,
        410,
    ]

    assert (unionarray + range(0, 100, 10)).tolist() == [
        0.0,
        11.1,
        22.2,
        30,
        140,
        53.3,
        64.4,
        270,
        380,
        490,
    ]
    assert (range(0, 100, 10) + unionarray).tolist() == [
        0.0,
        11.1,
        22.2,
        30,
        140,
        53.3,
        64.4,
        270,
        380,
        490,
    ]

    assert (unionarray + np.arange(0, 100, 10)).tolist() == [
        0.0,
        11.1,
        22.2,
        30,
        140,
        53.3,
        64.4,
        270,
        380,
        490,
    ]
    assert (np.arange(0, 100, 10) + unionarray).tolist() == [
        0.0,
        11.1,
        22.2,
        30,
        140,
        53.3,
        64.4,
        270,
        380,
        490,
    ]

    assert (unionarray + ak.Array(np.arange(0, 100, 10))).tolist() == [
        0.0,
        11.1,
        22.2,
        30,
        140,
        53.3,
        64.4,
        270,
        380,
        490,
    ]
    assert (ak.Array(np.arange(0, 100, 10)) + unionarray).tolist() == [
        0.0,
        11.1,
        22.2,
        30,
        140,
        53.3,
        64.4,
        270,
        380,
        490,
    ]

    assert (unionarray + unionarray).tolist() == [
        0.0,
        2.2,
        4.4,
        0,
        200,
        6.6,
        8.8,
        400,
        600,
        800,
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

from awkward._v2.tmp_for_testing import v1_to_v2

pytestmark = pytest.mark.skipif(
    ak._util.py27, reason="No Python 2.7 support in Awkward 2.x"
)


def test_basic():
    content = ak.layout.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4]))

    ind = np.array([2, 2, 0, 3, 4], dtype=np.int32)
    index = ak.layout.Index32(ind)
    array = ak.layout.IndexedArray32(index, content)

    array = v1_to_v2(array)

    assert ak.to_list(array) == [2.2, 2.2, 0.0, 3.3, 4.4]
    ind[3] = 1
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 1.1, 4.4]

    ind = np.array([2, 2, 0, 3, 4], dtype=np.uint32)
    index = ak.layout.IndexU32(ind)
    array = ak.layout.IndexedArrayU32(index, content)
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 3.3, 4.4]
    ind[3] = 1
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 1.1, 4.4]

    ind = np.array([2, 2, 0, 3, 4], dtype=np.int64)
    index = ak.layout.Index64(ind)
    array = ak.layout.IndexedArray64(index, content)
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 3.3, 4.4]
    ind[3] = 1
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 1.1, 4.4]

    ind = np.array([2, 2, 0, 3, 4], dtype=np.int32)
    index = ak.layout.Index32(ind)
    array = ak.layout.IndexedOptionArray32(index, content)
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 3.3, 4.4]
    ind[3] = 1
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 1.1, 4.4]

    ind = np.array([2, 2, 0, 3, 4], dtype=np.int64)
    index = ak.layout.Index64(ind)
    array = ak.layout.IndexedOptionArray64(index, content)
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 3.3, 4.4]
    ind[3] = 1
    assert ak.to_list(array) == [2.2, 2.2, 0.0, 1.1, 4.4]


def test_null():
    content = ak.layout.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4]))
    index = ak.layout.Index64(np.array([2, 2, 0, -1, 4], dtype=np.int64))
    array = ak.layout.IndexedOptionArray64(index, content)

    array = v1_to_v2(array)

    assert ak.to_list(array) == [2.2, 2.2, 0.0, None, 4.4]


def test_carry():
    content = ak.layout.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4]))
    index = ak.layout.Index64(np.array([2, 2, 0, 3, 4], dtype=np.int64))
    indexedarray = ak.layout.IndexedArray64(index, content)
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5], dtype=np.int64))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, indexedarray)

    listoffsetarray = v1_to_v2(listoffsetarray)

    assert ak.to_list(listoffsetarray) == [[2.2, 2.2, 0.0], [], [3.3, 4.4]]
    assert ak.to_list(listoffsetarray[::-1]) == [[3.3, 4.4], [], [2.2, 2.2, 0.0]]
    assert listoffsetarray.typetracer[::-1].form == listoffsetarray[::-1].form
    assert ak.to_list(listoffsetarray[[2, 0]]) == [[3.3, 4.4], [2.2, 2.2, 0.0]]
    assert listoffsetarray.typetracer[[2, 0]].form == listoffsetarray[[2, 0]].form
    assert ak.to_list(listoffsetarray[[2, 0], 1]) == [4.4, 2.2]  # invokes carry
    assert listoffsetarray.typetracer[[2, 0], 1].form == listoffsetarray[[2, 0], 1].form
    assert ak.to_list(listoffsetarray[2:, 1]) == [4.4]  # invokes carry
    assert listoffsetarray.typetracer[2:, 1].form == listoffsetarray[2:, 1].form

    index = ak.layout.Index64(np.array([2, 2, 0, 3, -1], dtype=np.int64))
    indexedarray = ak.layout.IndexedOptionArray64(index, content)
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, indexedarray)
    assert ak.to_list(listoffsetarray) == [[2.2, 2.2, 0.0], [], [3.3, None]]
    assert ak.to_list(listoffsetarray[::-1]) == [[3.3, None], [], [2.2, 2.2, 0.0]]
    assert ak.to_list(listoffsetarray[[2, 0]]) == [[3.3, None], [2.2, 2.2, 0.0]]
    assert ak.to_list(listoffsetarray[[2, 0], 1]) == [None, 2.2]  # invokes carry
    assert ak.to_list(listoffsetarray[2:, 1]) == [None]  # invokes carry


def test_others():
    content = ak.layout.NumpyArray(
        np.array(
            [[0.0, 0.0], [0.1, 1.0], [0.2, 2.0], [0.3, 3.0], [0.4, 4.0], [0.5, 5.0]]
        )
    )
    index = ak.layout.Index64(np.array([4, 0, 3, 1, 3], dtype=np.int64))
    indexedarray = ak.layout.IndexedArray64(index, content)

    indexedarray = v1_to_v2(indexedarray)

    assert indexedarray[3, 0] == 0.1
    assert indexedarray[3, 1] == 1.0
    assert ak.to_list(indexedarray[3, ::-1]) == [1.0, 0.1]
    assert indexedarray.typetracer[3, ::-1].form == indexedarray[3, ::-1].form
    assert ak.to_list(indexedarray[3, [1, 1, 0]]) == [1.0, 1.0, 0.1]
    assert indexedarray.typetracer[3, [1, 1, 0]].form == indexedarray[3, [1, 1, 0]].form
    assert ak.to_list(indexedarray[3:, 0]) == [0.1, 0.3]
    assert indexedarray.typetracer[3:, 0].form == indexedarray[3:, 0].form
    assert ak.to_list(indexedarray[3:, 1]) == [1.0, 3.0]
    assert indexedarray.typetracer[3:, 1].form == indexedarray[3:, 1].form
    assert ak.to_list(indexedarray[3:, ::-1]) == [[1.0, 0.1], [3.0, 0.3]]
    assert indexedarray.typetracer[3:, ::-1].form == indexedarray[3:, ::-1].form
    assert ak.to_list(indexedarray[3:, [1, 1, 0]]) == [[1.0, 1.0, 0.1], [3.0, 3.0, 0.3]]
    assert (
        indexedarray.typetracer[3:, [1, 1, 0]].form == indexedarray[3:, [1, 1, 0]].form
    )


def test_missing():
    content = ak.layout.NumpyArray(
        np.array(
            [[0.0, 0.0], [0.1, 1.0], [0.2, 2.0], [0.3, 3.0], [0.4, 4.0], [0.5, 5.0]]
        )
    )
    index = ak.layout.Index64(np.array([4, 0, 3, -1, 3], dtype=np.int64))
    indexedarray = ak.layout.IndexedOptionArray64(index, content)

    indexedarray = v1_to_v2(indexedarray)
    assert ak.to_list(indexedarray[3:, 0]) == [None, 0.3]
    assert indexedarray.typetracer[3:, 0].form == indexedarray[3:, 0].form
    assert ak.to_list(indexedarray[3:, 1]) == [None, 3.0]
    assert indexedarray.typetracer[3:, 1].form == indexedarray[3:, 1].form
    assert ak.to_list(indexedarray[3:, ::-1]) == [None, [3.0, 0.3]]
    assert indexedarray.typetracer[3:, ::-1].form == indexedarray[3:, ::-1].form
    assert ak.to_list(indexedarray[3:, [1, 1, 0]]) == [None, [3.0, 3.0, 0.3]]
    assert (
        indexedarray.typetracer[3:, [1, 1, 0]].form == indexedarray[3:, [1, 1, 0]].form
    )



import awkward as ak
import numpy as np
import pytest

jax = pytest.importorskip("jax")
jax.config.update("jax_platform_name", "cpu")
ak.jax.register()

# #### ak.layout.NumpyArray ####

test_numpyarray = ak.Array(np.arange(10, dtype=np.float64))
test_numpyarray_tangent = ak.Array(np.arange(10, dtype=np.float64))

test_numpyarray_jax = jax.numpy.arange(10, dtype=np.float64)
test_numpyarray_tangent_jax = jax.numpy.arange(10, dtype=np.float64)


def test_numpyarray_grad_1():
    def func_numpyarray_1(x):
        return x[4] ** 2

    value_jvp, jvp_grad = jax.jvp(
        func_numpyarray_1, (test_numpyarray,), (test_numpyarray_tangent,)
    )
    jit_value = jax.jit(func_numpyarray_1)(test_numpyarray)
    # value_vjp, vjp_func = jax.vjp(func_numpyarray_1, test_numpyarray)

    assert value_jvp == 16.0
    # assert value_vjp == 16.0
    assert jit_value == 16.0
    assert jvp_grad == 32.0
    # assert vjp_func(value_vjp)[0] == [0.0, 10.0, 72.0, 228.0, 200.0, 0.0, 0.0, 0.0, 0.0, 0.0]


def test_numpyarray_grad_2():
    def func_numpyarray_2(x):
        return x[2:5] ** 2 + x[1:4] ** 2

    value_jvp, jvp_grad = jax.jvp(
        func_numpyarray_2, (test_numpyarray,), (test_numpyarray_tangent,)
    )
    jit_value = jax.jit(func_numpyarray_2)(test_numpyarray)
    value_vjp, vjp_func = jax.vjp(func_numpyarray_2, test_numpyarray)

    assert ak.to_list(value_jvp) == [5, 13, 25]
    assert ak.to_list(value_vjp) == [5, 13, 25]
    assert ak.to_list(jit_value) == [5, 13, 25]
    assert ak.to_list(jvp_grad) == [10.0, 26.0, 50.0]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [
        0.0,
        10.0,
        72.0,
        228.0,
        200.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
    ]


def test_numpyarray_grad_3():
    def func_numpyarray_3(x):
        return x[::-1]

    value_jvp, jvp_grad = jax.jvp(
        func_numpyarray_3, (test_numpyarray_jax,), (test_numpyarray_tangent_jax,)
    )
    jit_value = jax.jit(func_numpyarray_3)(test_numpyarray)
    value_vjp, vjp_func = jax.vjp(func_numpyarray_3, test_numpyarray_jax)

    assert ak.to_list(value_jvp) == [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]
    assert ak.to_list(value_vjp) == [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]
    assert ak.to_list(jit_value) == [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]
    assert ak.to_list(jvp_grad) == [9.0, 8.0, 7.0, 6.0, 5.0, 4.0, 3.0, 2.0, 1.0, 0.0]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [
        0.0,
        1.0,
        2.0,
        3.0,
        4.0,
        5.0,
        6.0,
        7.0,
        8.0,
        9.0,
    ]


def test_numpyarray_grad_4():
    def func_numpyarray_4(x):
        return x[2:5] ** 2 * x[1:4] ** 2

    value_jvp, jvp_grad = jax.jvp(
        func_numpyarray_4, (test_numpyarray,), (test_numpyarray_tangent,)
    )
    jit_value = jax.jit(func_numpyarray_4)(test_numpyarray)
    value_vjp, vjp_func = jax.vjp(func_numpyarray_4, test_numpyarray)

    assert ak.to_list(value_jvp) == [4.0, 36.0, 144.0]
    assert ak.to_list(value_vjp) == [4.0, 36.0, 144.0]
    assert ak.to_list(jit_value) == [4.0, 36.0, 144.0]
    assert ak.to_list(jvp_grad) == [16.0, 144.0, 576.0]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [
        0.0,
        32.0,
        1312.0,
        14688.0,
        10368.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
    ]


test_listoffsetarray = ak.Array([[1.0, 2.0, 3.0], [], [4.0, 5.0]])
test_listoffsetarray_tangent = ak.Array([[0.0, 0.0, 0.0], [], [0.0, 1.0]])


def test_listoffset_array_1():
    def func_listoffsetarray_1(x):
        return x[2] * 2

    value_jvp, jvp_grad = jax.jvp(
        func_listoffsetarray_1, (test_listoffsetarray,), (test_listoffsetarray_tangent,)
    )
    jit_value = jax.jit(func_listoffsetarray_1)(test_listoffsetarray)
    value_vjp, vjp_func = jax.vjp(func_listoffsetarray_1, test_listoffsetarray)

    assert ak.to_list(value_jvp) == [8, 10]
    assert ak.to_list(value_vjp) == [8, 10]
    assert ak.to_list(jit_value) == [8, 10]
    assert ak.to_list(jvp_grad) == [0.0, 2.0]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [[0.0, 0.0, 0.0], [], [16.0, 20.0]]


def test_listoffset_array_2():
    def func_listoffsetarray_2(x):
        return x * x

    value_jvp, jvp_grad = jax.jvp(
        func_listoffsetarray_2, (test_listoffsetarray,), (test_listoffsetarray_tangent,)
    )
    jit_value = jax.jit(func_listoffsetarray_2)(test_listoffsetarray)
    value_vjp, vjp_func = jax.vjp(func_listoffsetarray_2, test_listoffsetarray)

    assert ak.to_list(value_jvp) == [[1.0, 4.0, 9.0], [], [16.0, 25.0]]
    assert ak.to_list(value_vjp) == [[1.0, 4.0, 9.0], [], [16.0, 25.0]]
    assert ak.to_list(jit_value) == [[1.0, 4.0, 9.0], [], [16.0, 25.0]]
    assert ak.to_list(jvp_grad) == [[0.0, 0.0, 0.0], [], [0.0, 10.0]]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [[2.0, 16.0, 54.0], [], [128.0, 250.0]]


def test_listoffset_array_3():
    def func_listoffsetarray_3(x):
        return x[0, 0] * x[2, 1]

    value_jvp, jvp_grad = jax.jvp(
        func_listoffsetarray_3, (test_listoffsetarray,), (test_listoffsetarray_tangent,)
    )
    jit_value = jax.jit(func_listoffsetarray_3)(test_listoffsetarray)
    # value_vjp, vjp_func = jax.vjp(func_listoffsetarray_2, test_listoffsetarray)

    assert ak.to_list(value_jvp) == 5.0
    # assert ak.to_list(value_vjp) == 2.0
    assert ak.to_list(jit_value) == 5.0
    assert ak.to_list(jvp_grad) == 1.0
    # assert ak.to_list(vjp_func(value_vjp)[0]) == [[2.0, 16.0, 54.0], [], [128.0, 250.0]]


def test_listoffset_array_4():
    def func_listoffsetarray_4(x):
        return x[::-1] ** 2

    value_jvp, jvp_grad = jax.jvp(
        func_listoffsetarray_4, (test_listoffsetarray,), (test_listoffsetarray_tangent,)
    )
    jit_value = jax.jit(func_listoffsetarray_4)(test_listoffsetarray)
    value_vjp, vjp_func = jax.vjp(func_listoffsetarray_4, test_listoffsetarray)

    assert ak.to_list(value_jvp) == [[16.0, 25.0], [], [1.0, 4.0, 9.0]]
    assert ak.to_list(value_vjp) == [[16.0, 25.0], [], [1.0, 4.0, 9.0]]
    assert ak.to_list(jit_value) == [[16.0, 25.0], [], [1.0, 4.0, 9.0]]
    assert ak.to_list(jvp_grad) == [[0.0, 10.0], [], [0.0, 0.0, 0.0]]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [[2.0, 16.0, 54.0], [], [128.0, 250.0]]


def test_listoffset_array_5():
    def func_listoffsetarray_5(x):
        return 2 * x[:-1]

    value_jvp, jvp_grad = jax.jvp(
        func_listoffsetarray_5, (test_listoffsetarray,), (test_listoffsetarray_tangent,)
    )
    jit_value = jax.jit(func_listoffsetarray_5)(test_listoffsetarray)
    value_vjp, vjp_func = jax.vjp(func_listoffsetarray_5, test_listoffsetarray)

    assert ak.to_list(value_jvp) == [[2.0, 4.0, 6.0], []]
    assert ak.to_list(value_vjp) == [[2.0, 4.0, 6.0], []]
    assert ak.to_list(jit_value) == [[2.0, 4.0, 6.0], []]
    assert ak.to_list(jvp_grad) == [[0.0, 0.0, 0.0], []]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [[4.0, 8.0, 12.0], [], [16.0, 20.0]]


def test_listoffset_array_6():
    def func_listoffsetarray_6(x):
        return x[0][0] * x[2][1]

    value_jvp, jvp_grad = jax.jvp(
        func_listoffsetarray_6, (test_listoffsetarray,), (test_listoffsetarray_tangent,)
    )
    jit_value = jax.jit(func_listoffsetarray_6)(test_listoffsetarray)
    # value_vjp, vjp_func = jax.vjp(func_listoffsetarray_2, test_listoffsetarray)

    assert ak.to_list(value_jvp) == 5.0
    # assert ak.to_list(value_vjp) == 2.0
    assert ak.to_list(jit_value) == 5.0
    assert ak.to_list(jvp_grad) == 1.0
    # assert ak.to_list(vjp_func(value_vjp)[0]) == [[2.0, 16.0, 54.0], [], [128.0, 250.0]]


test_recordarray = ak.Array(
    [
        [{"x": 1.1, "y": [1.0]}, {"x": 2.2, "y": [1.0, 2.2]}],
        [],
        [{"x": 3.3, "y": [1.0, 2.0, 3.0]}],
    ]
)
test_recordarray_tangent = ak.Array(
    [
        [{"x": 0.0, "y": [1.0]}, {"x": 2.0, "y": [1.5, 0.0]}],
        [],
        [{"x": 1.5, "y": [2.0, 0.5, 1.0]}],
    ]
)


def test_recordarray_1():
    def func_recordarray_1(x):
        return 2 * x.y[2][0][1] + 10

    value_jvp, jvp_grad = jax.jvp(
        func_recordarray_1, (test_recordarray,), (test_recordarray_tangent,)
    )
    jit_value = jax.jit(func_recordarray_1)(test_recordarray)
    # value_vjp, vjp_func = jax.vjp(func_listoffsetarray_2, test_listoffsetarray)

    assert ak.to_list(value_jvp) == 14.0
    # assert ak.to_list(value_vjp) == 14.0
    assert ak.to_list(jit_value) == 14.0
    assert ak.to_list(jvp_grad) == 1.0
    # assert ak.to_list(vjp_func(value_vjp)[0]) == [[2.0, 16.0, 54.0], [], [128.0, 250.0]]


def test_recordarray_2():
    def func_recordarray_2(x):
        return 2 * x.y[2][0] + 10

    value_jvp, jvp_grad = jax.jvp(
        func_recordarray_2, (test_recordarray,), (test_recordarray_tangent,)
    )
    jit_value = jax.jit(func_recordarray_2)(test_recordarray)
    value_vjp, vjp_func = jax.vjp(func_recordarray_2, test_recordarray)

    assert ak.to_list(value_jvp) == [12.0, 14.0, 16.0]
    assert ak.to_list(value_vjp) == [12.0, 14.0, 16.0]
    assert ak.to_list(jit_value) == [12.0, 14.0, 16.0]
    assert ak.to_list(jvp_grad) == [4.0, 1.0, 2.0]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [
        [{"x": 0.0, "y": [0.0]}, {"x": 0.0, "y": [0.0, 0.0]}],
        [],
        [{"x": 0.0, "y": [24.0, 28.0, 32.0]}],
    ]


def test_recordarray_3():
    def func_recordarray_3(x):
        return 2 * x.y[0][0] ** 2

    value_jvp, jvp_grad = jax.jvp(
        func_recordarray_3, (test_recordarray,), (test_recordarray_tangent,)
    )
    jit_value = jax.jit(func_recordarray_3)(test_recordarray)
    value_vjp, vjp_func = jax.vjp(func_recordarray_3, test_recordarray)

    assert ak.to_list(value_jvp) == [2.0]
    assert ak.to_list(value_vjp) == [2.0]
    assert ak.to_list(jit_value) == [2.0]
    assert ak.to_list(jvp_grad) == [4.0]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [
        [{"x": 0.0, "y": [8.0]}, {"x": 0.0, "y": [0.0, 0.0]}],
        [],
        [{"x": 0.0, "y": [0.0, 0.0, 0.0]}],
    ]


def test_recordarray_4():
    def func_recordarray_4(x):
        return 2 * x.y[2] + 10

    value_jvp, jvp_grad = jax.jvp(
        func_recordarray_4, (test_recordarray,), (test_recordarray_tangent,)
    )
    jit_value = jax.jit(func_recordarray_4)(test_recordarray)
    value_vjp, vjp_func = jax.vjp(func_recordarray_4, test_recordarray)

    assert ak.to_list(value_jvp) == [[12.0, 14.0, 16.0]]
    assert ak.to_list(value_vjp) == [[12.0, 14.0, 16.0]]
    assert ak.to_list(jit_value) == [[12.0, 14.0, 16.0]]
    assert ak.to_list(jvp_grad) == [[4.0, 1.0, 2.0]]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [
        [{"x": 0.0, "y": [24.0]}, {"x": 0.0, "y": [24.0, 28.8]}],
        [],
        [{"x": 0.0, "y": [24.0, 28.0, 32.0]}],
    ]


def test_recordarray_5():
    def func_recordarray_5(x):
        return 2 * x.y

    value_jvp, jvp_grad = jax.jvp(
        func_recordarray_5, (test_recordarray,), (test_recordarray_tangent,)
    )
    jit_value = jax.jit(func_recordarray_5)(test_recordarray)
    value_vjp, vjp_func = jax.vjp(func_recordarray_5, test_recordarray)

    assert ak.to_list(value_jvp) == [[[2.0], [2.0, 4.4]], [], [[2.0, 4.0, 6.0]]]
    assert ak.to_list(value_vjp) == [[[2.0], [2.0, 4.4]], [], [[2.0, 4.0, 6.0]]]
    assert ak.to_list(jit_value) == [[[2.0], [2.0, 4.4]], [], [[2.0, 4.0, 6.0]]]
    assert ak.to_list(jvp_grad) == [[[2.0], [3.0, 0.0]], [], [[4.0, 1.0, 2.0]]]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [
        [{"x": 0.0, "y": [4.0]}, {"x": 0.0, "y": [4.0, 8.8]}],
        [],
        [{"x": 0.0, "y": [4.0, 8.0, 12.0]}],
    ]


def test_recordarray_6():
    def func_recordarray_6(x):
        return 2 * x.y ** 2

    value_jvp, jvp_grad = jax.jvp(
        func_recordarray_6, (test_recordarray,), (test_recordarray_tangent,)
    )
    jit_value = jax.jit(func_recordarray_6)(test_recordarray)
    value_vjp, vjp_func = jax.vjp(func_recordarray_6, test_recordarray)

    assert ak.to_list(value_jvp) == [
        [[2.0], [2.0, 9.680000000000001]],
        [],
        [[2.0, 8.0, 18.0]],
    ]
    assert ak.to_list(value_vjp) == [
        [[2.0], [2.0, 9.680000000000001]],
        [],
        [[2.0, 8.0, 18.0]],
    ]
    assert ak.to_list(jit_value) == [
        [[2.0], [2.0, 9.680000000000001]],
        [],
        [[2.0, 8.0, 18.0]],
    ]
    assert ak.to_list(jvp_grad) == [[[4.0], [6.0, 0.0]], [], [[8.0, 4.0, 12.0]]]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [
        [{"x": 0.0, "y": [8.0]}, {"x": 0.0, "y": [8.0, 85.18400000000003]}],
        [],
        [{"x": 0.0, "y": [8.0, 64.0, 216.0]}],
    ]


def test_recordarray_7():
    def func_recordarray_7(x):
        return 2 * x.y[2, 0, 1] + 10

    value_jvp, jvp_grad = jax.jvp(
        func_recordarray_7, (test_recordarray,), (test_recordarray_tangent,)
    )
    jit_value = jax.jit(func_recordarray_7)(test_recordarray)
    # value_vjp, vjp_func = jax.vjp(func_listoffsetarray_2, test_listoffsetarray)

    assert ak.to_list(value_jvp) == 14.0
    # assert ak.to_list(value_vjp) == 14.0
    assert ak.to_list(jit_value) == 14.0
    assert ak.to_list(jvp_grad) == 1.0
    # assert ak.to_list(vjp_func(value_vjp)[0]) == [[2.0, 16.0, 54.0], [], [128.0, 250.0]]


def test_recordarray_8():
    def func_recordarray_8(x):
        return 2 * x.y[2, 0] + 10

    value_jvp, jvp_grad = jax.jvp(
        func_recordarray_8, (test_recordarray,), (test_recordarray_tangent,)
    )
    jit_value = jax.jit(func_recordarray_8)(test_recordarray)
    value_vjp, vjp_func = jax.vjp(func_recordarray_8, test_recordarray)

    assert ak.to_list(value_jvp) == [12.0, 14.0, 16.0]
    assert ak.to_list(value_vjp) == [12.0, 14.0, 16.0]
    assert ak.to_list(jit_value) == [12.0, 14.0, 16.0]
    assert ak.to_list(jvp_grad) == [4.0, 1.0, 2.0]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [
        [{"x": 0.0, "y": [0.0]}, {"x": 0.0, "y": [0.0, 0.0]}],
        [],
        [{"x": 0.0, "y": [24.0, 28.0, 32.0]}],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import json

import awkward as ak
from awkward._v2.index import Index
from awkward._v2._slicing import NestedIndexError
from awkward._v2.contents.content import Content
from awkward._v2.contents.bytemaskedarray import ByteMaskedArray
from awkward._v2.forms.bitmaskedform import BitMaskedForm

np = ak.nplike.NumpyMetadata.instance()


class BitMaskedArray(Content):
    def __init__(
        self,
        mask,
        content,
        valid_when,
        length,
        lsb_order,
        identifier=None,
        parameters=None,
    ):
        if not (isinstance(mask, Index) and mask.dtype == np.dtype(np.uint8)):
            raise TypeError(
                "{0} 'mask' must be an Index with dtype=uint8, not {1}".format(
                    type(self).__name__, repr(mask)
                )
            )
        if not isinstance(content, Content):
            raise TypeError(
                "{0} 'content' must be a Content subtype, not {1}".format(
                    type(self).__name__, repr(content)
                )
            )
        if not isinstance(valid_when, bool):
            raise TypeError(
                "{0} 'valid_when' must be boolean, not {1}".format(
                    type(self).__name__, repr(valid_when)
                )
            )
        if not ak._util.isint(length):
            raise TypeError(
                "{0} 'length' must be an integer, not {1}".format(
                    type(self).__name__, repr(length)
                )
            )
        if not isinstance(lsb_order, bool):
            raise TypeError(
                "{0} 'lsb_order' must be boolean, not {1}".format(
                    type(self).__name__, repr(lsb_order)
                )
            )
        if not length <= len(mask) * 8:
            raise ValueError(
                "{0} 'length' ({1}) must be <= len(mask) * 8 ({2})".format(
                    type(self).__name__, length, len(mask) * 8
                )
            )
        if not length <= len(content):
            raise ValueError(
                "{0} 'length' ({1}) must be <= len(content) ({2})".format(
                    type(self).__name__, length, len(content)
                )
            )

        self._mask = mask
        self._content = content
        self._valid_when = valid_when
        self._length = length
        self._lsb_order = lsb_order
        self._init(identifier, parameters)

    @property
    def mask(self):
        return self._mask

    @property
    def content(self):
        return self._content

    @property
    def valid_when(self):
        return self._valid_when

    @property
    def lsb_order(self):
        return self._lsb_order

    @property
    def nplike(self):
        return self._mask.nplike

    @property
    def nonvirtual_nplike(self):
        return self._mask.nplike

    Form = BitMaskedForm

    @property
    def form(self):
        return self.Form(
            self._mask.form,
            self._content.form,
            self._valid_when,
            self._lsb_order,
            has_identifier=self._identifier is not None,
            parameters=self._parameters,
            form_key=None,
        )

    @property
    def typetracer(self):
        return BitMaskedArray(
            ak._v2.index.Index(self._mask.to(ak._v2._typetracer.TypeTracer.instance())),
            self._content.typetracer,
            self._valid_when,
            self._length,
            self._lsb_order,
            self._typetracer_identifier(),
            self._parameters,
        )

    def __len__(self):
        return self._length

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<BitMaskedArray valid_when="]
        out.append(repr(json.dumps(self._valid_when)))
        out.append(" lsb_order=")
        out.append(repr(json.dumps(self._lsb_order)))
        out.append(" len=")
        out.append(repr(str(len(self))))
        out.append(">")
        out.extend(self._repr_extra(indent + "    "))
        out.append("\n")
        out.append(self._mask._repr(indent + "    ", "<mask>", "</mask>\n"))
        out.append(self._content._repr(indent + "    ", "<content>", "</content>\n"))
        out.append(indent + "</BitMaskedArray>")
        out.append(post)
        return "".join(out)

    def toByteMaskedArray(self):
        nplike = self._mask.nplike
        bytemask = ak._v2.index.Index8.empty(len(self._mask) * 8, nplike)
        self._handle_error(
            nplike[
                "awkward_BitMaskedArray_to_ByteMaskedArray",
                bytemask.dtype.type,
                self._mask.dtype.type,
            ](
                bytemask.to(nplike),
                self._mask.to(nplike),
                len(self._mask),
                False,  # this differs from the kernel call in 'bytemask'
                self._lsb_order,
            )
        )
        return ByteMaskedArray(
            bytemask[: self._length],
            self._content,
            self._valid_when,
            self._identifier,
            self._parameters,
        )

    def toIndexedOptionArray64(self):
        nplike = self._mask.nplike
        index = ak._v2.index.Index64.empty(len(self._mask) * 8, nplike)
        self._handle_error(
            nplike[
                "awkward_BitMaskedArray_to_IndexedOptionArray",
                index.dtype.type,
                self._mask.dtype.type,
            ](
                index.to(nplike),
                self._mask.to(nplike),
                len(self._mask),
                self._valid_when,
                self._lsb_order,
            ),
        )
        return ak._v2.contents.indexedoptionarray.IndexedOptionArray(
            index[0 : self._length],
            self._content,
            self._identifier,
            self._parameters,
        )

    def mask_as_bool(self, valid_when=None):
        if valid_when is None:
            valid_when = self._valid_when
        nplike = self._mask.nplike
        bytemask = ak._v2.index.Index8.empty(len(self._mask) * 8, nplike)
        self._handle_error(
            nplike[
                "awkward_BitMaskedArray_to_ByteMaskedArray",
                bytemask.dtype.type,
                self._mask.dtype.type,
            ](
                bytemask.to(nplike),
                self._mask.to(nplike),
                len(self._mask),
                0 if valid_when else 1,
                self._lsb_order,
            )
        )
        return bytemask.data[: self._length].view(np.bool_)

    def _getitem_nothing(self):
        return self._content._getitem_range(slice(0, 0))

    def _getitem_at(self, where):
        if where < 0:
            where += len(self)
        if not (0 <= where < len(self)) and self.nplike.known_shape:
            raise NestedIndexError(self, where)
        if self._lsb_order:
            bit = bool(self._mask[where // 8] & (1 << (where % 8)))
        else:
            bit = bool(self._mask[where // 8] & (128 >> (where % 8)))
        if bit == self._valid_when:
            return self._content._getitem_at(where)
        else:
            return None

    def _getitem_range(self, where):
        return self.toByteMaskedArray()._getitem_range(where)

    def _getitem_field(self, where, only_fields=()):
        return BitMaskedArray(
            self._mask,
            self._content._getitem_field(where, only_fields),
            self._valid_when,
            self._length,
            self._lsb_order,
            self._field_identifier(where),
            None,
        )

    def _getitem_fields(self, where, only_fields=()):
        return BitMaskedArray(
            self._mask,
            self._content._getitem_fields(where, only_fields),
            self._valid_when,
            self._length,
            self._lsb_order,
            self._fields_identifier(where),
            None,
        )

    def _carry(self, carry, allow_lazy, exception):
        assert isinstance(carry, ak._v2.index.Index)
        return self.toByteMaskedArray()._carry(carry, allow_lazy, exception)

    def _getitem_next_jagged(self, slicestarts, slicestops, slicecontent, tail):
        return self.toByteMaskedArray()._getitem_next_jagged(
            slicestarts, slicestops, slicecontent, tail
        )

    def _getitem_next(self, head, tail, advanced):
        if head == ():
            return self

        elif isinstance(head, (int, slice, ak._v2.index.Index64)):
            return self.toByteMaskedArray()._getitem_next(head, tail, advanced)

        elif ak._util.isstr(head):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif head is np.newaxis:
            return self._getitem_next_newaxis(tail, advanced)

        elif head is Ellipsis:
            return self._getitem_next_ellipsis(tail, advanced)

        elif isinstance(head, ak._v2.contents.ListOffsetArray):
            return self.toByteMaskedArray()._getitem_next(head, tail, advanced)

        elif isinstance(head, ak._v2.contents.IndexedOptionArray):
            return self._getitem_next_missing(head, tail, advanced)

        else:
            raise AssertionError(repr(head))

    def _localindex(self, axis, depth):
        return self.toByteMaskedArray()._localindex(axis, depth)

    def _argsort_next(
        self,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        ascending,
        stable,
        kind,
        order,
    ):
        if len(self._mask) == 0:
            return ak._v2.contents.NumpyArray(self.nplike.empty(0, np.int64))

        return self.toIndexedOptionArray64()._argsort_next(
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )

    def _sort_next(
        self, negaxis, starts, parents, outlength, ascending, stable, kind, order
    ):
        return self.toIndexedOptionArray64()._sort_next(
            negaxis,
            starts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        return self.toByteMaskedArray()._combinations(
            n, replacement, recordlookup, parameters, axis, depth
        )

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
    ):
        return self.toByteMaskedArray()._reduce_next(
            reducer,
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            mask,
            keepdims,
        )

    def _validityerror(self, path):
        if len(self.mask) * 8 < len(self):
            return 'at {0} ("{1}"): len(mask) * 8 < length'.format(path, type(self))
        elif len(self.content) < len(self):
            return 'at {0} ("{1}"): len(content) < length'.format(path, type(self))
        elif isinstance(
            self.content,
            (
                ak._v2.contents.bitmaskedarray.BitMaskedArray,
                ak._v2.contents.bytemaskedarray.ByteMaskedArray,
                ak._v2.contents.indexedarray.IndexedArray,
                ak._v2.contents.indexedoptionarray.IndexedOptionArray,
                ak._v2.contents.unmaskedarray.UnmaskedArray,
            ),
        ):
            return "{0} contains \"{1}\", the operation that made it might have forgotten to call 'simplify_optiontype()'"
        else:
            return self.content.validityerror(path + ".content")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

import awkward.forth


def test_record_form2():
    form = """
{
    "class": "RecordArray",
    "contents": {
        "one": "float64",
        "two": "int64"
    },
    "form_key": "node0"
}
    """
    builder = ak.layout.LayoutBuilder32(form)

    # the fields alternate
    builder.float64(1.1)  # "one"
    builder.int64(2)  # "two"
    builder.float64(3.3)  # "one"
    builder.int64(4)  # "two"

    # etc.

    assert ak.to_list(builder.snapshot()) == [
        {"one": 1.1, "two": 2},
        {"one": 3.3, "two": 4},
    ]


def test_bit_masked_form():
    form = """
{
    "class": "BitMaskedArray",
    "mask": "i8",
    "content": "float64",
    "valid_when": true,
    "lsb_order": false,
    "form_key": "node0"
}
    """
    builder = ak.layout.LayoutBuilder32(form)

    builder.float64(1.1)
    builder.float64(2.2)
    builder.float64(3.3)

    assert ak.to_list(builder.snapshot()) == [1.1, 2.2, 3.3]


def test_byte_masked_form():
    form = """
{
    "class": "ByteMaskedArray",
    "mask": "i8",
    "content": "float64",
    "valid_when": true,
    "form_key": "node0"
}
    """

    builder = ak.layout.LayoutBuilder32(form)

    builder.float64(1.1)
    builder.float64(2.2)
    builder.float64(3.3)

    assert ak.to_list(builder.snapshot()) == [1.1, 2.2, 3.3]


def test_unmasked_form():
    form = """
{
    "class": "UnmaskedArray",
    "content": "float64",
    "form_key": "node0"
}
"""
    builder = ak.layout.LayoutBuilder32(form)

    builder.float64(1.1)
    builder.float64(2.2)
    builder.float64(3.3)

    assert ak.to_list(builder.snapshot()) == [1.1, 2.2, 3.3]


def test_unsupported_form():
    form = """
{
    "class": "VirtualArray",
    "form": "float64",
    "has_length": true
}
           """

    with pytest.raises(ValueError):
        ak.layout.LayoutBuilder32(form)


def test_list_offset_form():
    form = """
{
    "class": "ListOffsetArray64",
    "offsets": "i64",
    "content": {
        "class": "RecordArray",
        "contents": {
            "x": {
                "class": "NumpyArray",
                "primitive": "float64",
                "form_key": "node2"
            },
            "y": {
                "class": "ListOffsetArray64",
                "offsets": "i64",
                "content": {
                    "class": "NumpyArray",
                    "primitive": "int64",
                    "form_key": "node4"
                },
                "form_key": "node3"
            }
        },
        "form_key": "node1"
    },
    "form_key": "node0"
}
    """

    builder = ak.layout.LayoutBuilder32(form)

    builder.begin_list()
    builder.float64(1.1)
    builder.begin_list()
    builder.int64(1)
    builder.end_list()
    builder.float64(2.2)
    builder.begin_list()
    builder.int64(1)
    builder.int64(2)
    builder.end_list()
    builder.end_list()

    builder.begin_list()
    builder.end_list()

    builder.begin_list()
    builder.float64(3.3)
    builder.begin_list()
    builder.int64(1)
    builder.int64(2)
    builder.int64(3)
    builder.end_list()
    builder.end_list()

    assert ak.to_list(builder.snapshot()) == [
        [{"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}],
        [],
        [{"x": 3.3, "y": [1, 2, 3]}],
    ]


def test_indexed_form():
    form = """
{
    "class": "IndexedArray64",
    "index": "i64",
    "content": {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "l",
        "primitive": "int64",
        "form_key": "node1"
    },
    "form_key": "node0"
}
    """

    builder = ak.layout.LayoutBuilder32(form)

    builder.int64(11)
    builder.int64(22)
    builder.int64(33)
    builder.int64(44)
    builder.int64(55)
    builder.int64(66)
    builder.int64(77)

    assert ak.to_list(builder.snapshot()) == [11, 22, 33, 44, 55, 66, 77]


def test_indexed_option_form():
    form = """
{
    "class": "IndexedOptionArray64",
    "index": "i64",
    "content": {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "l",
        "primitive": "int64",
        "form_key": "node1"
    },
    "form_key": "node0"
}
    """

    builder = ak.layout.LayoutBuilder32(form)

    builder.null()
    builder.int64(11)
    builder.int64(22)
    builder.null()
    builder.int64(33)
    builder.int64(44)
    builder.null()
    builder.int64(55)
    builder.int64(66)
    builder.int64(77)

    assert ak.to_list(builder.snapshot()) == [
        None,
        11,
        22,
        None,
        33,
        44,
        None,
        55,
        66,
        77,
    ]


def test_regular_form():
    form = """
{
    "class": "RegularArray",
    "content": {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "l",
        "primitive": "int64",
        "form_key": "node1"
    },
    "size": 3,
    "form_key": "node0"
}
    """

    builder = ak.layout.LayoutBuilder32(form)

    builder.int64(11)
    builder.int64(22)
    builder.int64(33)
    builder.int64(44)
    builder.int64(55)
    builder.int64(66)
    builder.int64(77)

    assert ak.to_list(builder.snapshot()) == [[11, 22, 33], [44, 55, 66]]

    builder.int64(88)
    builder.int64(99)

    assert ak.to_list(builder.snapshot()) == [[11, 22, 33], [44, 55, 66], [77, 88, 99]]


def test_union_form():
    form = """
{
    "class": "UnionArray8_64",
    "tags": "i8",
    "index": "i64",
    "contents": [
        "float64",
        "bool"
    ],
    "form_key": "node0"
}
    """

    builder = ak.layout.LayoutBuilder32(form)

    builder.tag(0)
    builder.float64(1.1)
    builder.tag(1)
    builder.boolean(False)
    builder.tag(0)
    builder.float64(2.2)
    builder.tag(0)
    builder.float64(3.3)
    builder.tag(1)
    builder.boolean(True)
    builder.tag(0)
    builder.float64(4.4)
    builder.tag(1)
    builder.boolean(False)
    builder.tag(1)
    builder.boolean(True)
    builder.tag(0)
    builder.float64(-2.2)

    assert ak.to_list(builder.snapshot()) == [
        1.1,
        False,
        2.2,
        3.3,
        True,
        4.4,
        False,
        True,
        -2.2,
    ]


def test_union2_form():
    form = """
{
    "class": "UnionArray8_64",
    "tags": "i8",
    "index": "i64",
    "contents": [
        "float64",
        "float64"
    ],
    "form_key": "node0"
}
    """

    builder = ak.layout.LayoutBuilder32(form)

    builder.tag(0)
    builder.float64(1.1)

    builder.tag(1)
    builder.float64(2.2)

    builder.tag(1)
    builder.float64(3.3)

    assert ak.to_list(builder.snapshot()) == [1.1, 2.2, 3.3]


def test_union3_form():
    form = """
{
    "class": "UnionArray8_64",
    "tags": "i8",
    "index": "i64",
    "contents": [
        "float64",
        "bool",
        "int64"
    ],
    "form_key": "node0"
}
    """

    builder = ak.layout.LayoutBuilder32(form)

    builder.tag(0)
    builder.float64(1.1)
    builder.tag(1)
    builder.boolean(False)
    builder.tag(2)
    builder.int64(11)
    builder.tag(0)
    builder.float64(2.2)
    builder.tag(1)
    builder.boolean(False)
    builder.tag(0)
    builder.float64(2.2)
    builder.tag(0)
    builder.float64(3.3)
    builder.tag(1)
    builder.boolean(True)
    builder.tag(0)
    builder.float64(4.4)
    builder.tag(1)
    builder.boolean(False)
    builder.tag(1)
    builder.boolean(True)
    builder.tag(0)
    builder.float64(-2.2)

    assert ak.to_list(builder.snapshot()) == [
        1.1,
        False,
        11,
        2.2,
        False,
        2.2,
        3.3,
        True,
        4.4,
        False,
        True,
        -2.2,
    ]


def test_record_form():
    form = """
{
    "class": "RecordArray",
    "contents": {
        "one": "float64",
        "two": "float64"
    },
    "form_key": "node0"
}
    """
    builder = ak.layout.LayoutBuilder32(form)

    # if record contents have the same type,
    # the fields alternate
    builder.float64(1.1)  # "one"
    builder.float64(2.2)  # "two"
    builder.float64(3.3)  # "one"
    builder.float64(4.4)  # "two"

    # etc.

    assert ak.to_list(builder.snapshot()) == [
        {"one": 1.1, "two": 2.2},
        {"one": 3.3, "two": 4.4},
    ]


def test_error_in_record_form():
    form = """
{
    "class": "RecordArray",
    "contents": {
        "one": "float64",
        "two": "float64"
    },
    "form_key": "node0"
}
    """

    builder = ak.layout.LayoutBuilder32(form)

    # if record contents have the same type,
    # the fields alternate
    builder.float64(1.1)  # "one"
    builder.float64(2.2)  # "two"
    with pytest.raises(ValueError) as err:
        builder.int64(11)
    assert str(err.value) == "NumpyForm builder accepts only float64"


def test_error_in_numpy_form():
    form = """
{
    "class": "NumpyArray",
    "itemsize": 8,
    "format": "d",
    "primitive": "float64"
}
    """

    builder = ak.layout.LayoutBuilder32(form)

    builder.float64(1.1)
    builder.float64(2.2)
    with pytest.raises(ValueError) as err:
        builder.int64(11)
    assert str(err.value) == "NumpyForm builder accepts only float64"


def test_categorical_form():
    form = """
{
    "class": "IndexedArray64",
    "index": "i64",
    "content": "int64",
    "parameters": {
        "__array__": "categorical"
    }
}
"""

    builder = ak.layout.LayoutBuilder32(form)

    builder.int64(2019)
    builder.int64(2020)
    builder.int64(2021)
    builder.int64(2020)
    builder.int64(2019)
    builder.int64(2020)
    builder.int64(2020)
    builder.int64(2020)
    builder.int64(2020)
    builder.int64(2020)

    assert ak.to_list(builder.snapshot()) == [
        2019,
        2020,
        2021,
        2020,
        2019,
        2020,
        2020,
        2020,
        2020,
        2020,
    ]
    assert str(ak.type(ak.Array(builder.snapshot()))) == "10 * categorical[type=int64]"


def test_char_form():
    form = """
{
    "class": "NumpyArray",
    "itemsize": 1,
    "format": "B",
    "primitive": "uint8",
    "parameters": {
        "__array__": "char"
    }
}"""

    builder = ak.layout.LayoutBuilder32(form)

    builder.string("one")
    builder.string("two")
    builder.string("three")

    assert ak.to_list(builder.snapshot()) == "onetwothree"


def test_string_form():
    form = """
{
    "class": "ListOffsetArray64",
    "offsets": "i64",
    "content": {
        "class": "NumpyArray",
        "itemsize": 1,
        "format": "B",
        "primitive": "uint8",
        "parameters": {
            "__array__": "char"
        }
    },
    "parameters": {
        "__array__": "string"
    }
}"""

    builder = ak.layout.LayoutBuilder32(form)

    builder.string("one")
    builder.string("two")
    builder.string("three")

    assert ak.to_list(builder.snapshot()) == ["one", "two", "three"]


def test_empty_form():
    form = """{
           "class": "ListOffsetArray64",
           "offsets": "i64",
           "content": {
               "class": "ListOffsetArray64",
               "offsets": "i64",
               "content": {
                   "class": "EmptyArray"
               },
               "form_key": "node1"
           },
           "form_key": "node0"
       }"""

    builder = ak.layout.LayoutBuilder32(form)

    builder.begin_list()
    builder.end_list()
    builder.begin_list()
    builder.begin_list()
    builder.end_list()
    builder.begin_list()
    builder.end_list()
    builder.begin_list()
    builder.end_list()
    builder.end_list()
    builder.begin_list()
    builder.begin_list()
    builder.end_list()
    builder.begin_list()
    builder.end_list()
    builder.end_list()
    builder.begin_list()
    builder.end_list()
    builder.begin_list()
    builder.begin_list()
    # builder.int64(1)
    # will fail with ValueError: EmptyArrayBuilder does not accept 'int64'
    builder.end_list()
    builder.end_list()

    assert ak.to_list(builder.snapshot()) == [[], [[], [], []], [[], []], [], [[]]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import awkward as ak
from awkward._v2.forms.form import Form, _parameters_equal, nonvirtual


class UnionForm(Form):
    def __init__(
        self,
        tags,
        index,
        contents,
        has_identifier=False,
        parameters=None,
        form_key=None,
    ):
        if not ak._util.isstr(tags):
            raise TypeError(
                "{0} 'tags' must be of type str, not {1}".format(
                    type(self).__name__, repr(tags)
                )
            )
        if not ak._util.isstr(index):
            raise TypeError(
                "{0} 'index' must be of type str, not {1}".format(
                    type(self).__name__, repr(index)
                )
            )
        if not isinstance(contents, Iterable):
            raise TypeError(
                "{0} 'contents' must be iterable, not {1}".format(
                    type(self).__name__, repr(contents)
                )
            )
        for content in contents:
            if not isinstance(content, Form):
                raise TypeError(
                    "{0} all 'contents' must be Form subclasses, not {1}".format(
                        type(self).__name__, repr(content)
                    )
                )

        self._tags = tags
        self._index = index
        self._contents = list(contents)
        self._init(has_identifier, parameters, form_key)

    @property
    def tags(self):
        return self._tags

    @property
    def index(self):
        return self._index

    @property
    def contents(self):
        return self._contents

    def content(self, index):
        return self._contents[index]

    def __repr__(self):
        args = [
            repr(self._tags),
            repr(self._index),
            repr(self._contents),
        ] + self._repr_args()
        return "{0}({1})".format(type(self).__name__, ", ".join(args))

    def _tolist_part(self, verbose, toplevel):
        return self._tolist_extra(
            {
                "class": "UnionArray",
                "tags": self._tags,
                "index": self._index,
                "contents": [
                    content._tolist_part(verbose, toplevel=False)
                    for content in self._contents
                ],
            },
            verbose,
        )

    def _type(self, typestrs):
        return ak._v2.types.uniontype.UnionType(
            [x._type(typestrs) for x in self._contents],
            self._parameters,
            ak._util.gettypestr(self._parameters, typestrs),
        )

    def __eq__(self, other):
        if isinstance(other, UnionForm):
            if (
                self._has_identifier == other._has_identifier
                and self._form_key == other._form_key
                and self._tags == other._tags
                and self._index == other._index
                and len(self._contents) == len(other._contents)
                and _parameters_equal(self._parameters, other._parameters)
            ):
                for i in range(len(self._contents)):
                    if self._contents[i] != other._contents[i]:
                        return False
                else:
                    return True
            else:
                return False
        else:
            return False

    def generated_compatibility(self, other):
        other = nonvirtual(other)

        if other is None:
            return True

        elif isinstance(other, UnionForm):
            if len(self._contents) == len(other._contents):
                return _parameters_equal(self._parameters, other._parameters) and all(
                    x.generated_compatibility(y)
                    for x, y in zip(self._contents, other._contents)
                )
            else:
                return False

        else:
            return False

    def _getitem_range(self):
        return UnionForm(
            self._tags,
            self._index,
            self._contents,
            has_identifier=self._has_identifier,
            parameters=self._parameters,
            form_key=None,
        )

    def _getitem_field(self, where, only_fields=()):
        return UnionForm(
            self._tags,
            self._index,
            self._content._getitem_field(where, only_fields),
            has_identifier=self._has_identifier,
            parameters=None,
            form_key=None,
        )

    def _getitem_fields(self, where, only_fields=()):
        return UnionForm(
            self._tags,
            self._index,
            self._content._getitem_fields(where, only_fields),
            has_identifier=self._has_identifier,
            parameters=None,
            form_key=None,
        )

    def _carry(self, allow_lazy):
        return UnionForm(
            self._tags,
            self._index,
            self._contents,
            has_identifier=self._has_identifier,
            parameters=self._parameters,
            form_key=None,
        )

    @property
    def purelist_isregular(self):
        for content in self._contents:
            if not content.purelist_isregular:
                return False
        else:
            return True

    @property
    def purelist_depth(self):
        out = None
        for content in self._contents:
            if out is None:
                out = content.purelist_depth
            elif out != content.purelist_depth:
                return -1
        else:
            return out

    @property
    def minmax_depth(self):
        if len(self._contents) == 0:
            return (0, 0)
        mins, maxs = [], []
        for content in self._contents:
            mindepth, maxdepth = content.minmax_depth
            mins.append(mindepth)
            maxs.append(maxdepth)
        return (min(mins), max(maxs))

    @property
    def branch_depth(self):
        if len(self._contents) == 0:
            return (False, 1)
        anybranch = False
        mindepth = None
        for content in self._contents:
            branch, depth = content.branch_depth
            if mindepth is None:
                mindepth = depth
            if branch or mindepth != depth:
                anybranch = True
            if mindepth > depth:
                mindepth = depth
        return (anybranch, mindepth)

    @property
    def keys(self):
        keyslists = []
        for i in range(len(self._contents)):
            keyslists.append(self._contents[i].keys)
        return list(set.intersection(*[set(x) for x in keyslists]))

    @property
    def dimension_optiontype(self):
        for content in self._contents:
            if content.dimension_optiontype:
                return True
        return False



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import os

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


pyarrow = pytest.importorskip("pyarrow")


def test_issue1():
    null_array = ak.Array({"null_col": [None]})
    tpe = ak.to_arrow_table(null_array).to_batches()[0].field("null_col").type
    assert tpe == pyarrow.null()


pyarrow_parquet = pytest.importorskip("pyarrow.parquet")


def test_issue2(tmp_path):
    filename = os.path.join(tmp_path, "whatever.parquet")

    null_table = pyarrow.Table.from_pydict({"null_col": pyarrow.array([None])})
    pyarrow_parquet.write_table(null_table, filename)

    assert ak.from_parquet(filename).type == ak.from_parquet(filename, lazy=True).type
